#! /bin/bash

RED_COLOR="\E[31m"
GREEN_COLOR="\E[32m"
YELLOW_COLOR="\E[33m"
RESET_COLOR="\E[37m"

# store all arguments in ARGS. 
ARGV=("$@")
ARGC=${#ARGV[@]}

# the first argument is the representation matrix of the problem.
ROOT=${ARGV[0]}
DOC="${ROOT}/doc"
PLOT_FOLDER=${DOC}/gnuplot

BIN=${ARGV[1]}
REFERENCE_PROGRAM=( "four-ti-two" "skeleton" "ppl" "cddr" "polco" "porta" "addibit-bp-trees" "addibit-bp-qbn-trees" "addibit-vp-trees" "addibit-pst" "addibit-alg" )

TIME_LIMIT=${ARGV[2]}

TIME_PLOT="${DOC}/performance.dat"
MEM_PLOT="${DOC}/memory.dat"

CONFIGURATION=( ) 

declare -a OVERALL_TIME
declare -a OVERALL_MEM

PROBLEM_COUNTER=1

if [ ! -e ${DOC} ]; then
    mkdir ${DOC}
fi
if [ ! -e ${PLOT_FOLDER} ]; then
    mkdir ${PLOT_FOLDER}
fi

if [ -e ${TIME_PLOT} ]; then
    rm ${TIME_PLOT}
fi

if [ -e ${MEM_PLOT} ]; then
    rm ${MEM_PLOT}
fi

function traverse () {
    local FOLDER=$1

    for FILE in `ls -Sr ${FOLDER}/* | egrep '\.ine$|\.ver$'`; do
        if [ -d "$FILE" ]; then
            traverse "${FILE}"
        else
            if [ ${FILE: -4} == ".ine" ] || [ ${FILE: -4} == ".ver" ]; then
                TABLE="${TABLE}\t\t<tr>"
                run "${FILE}"
                TABLE="${TABLE}</tr>\n"
            fi
        fi
    done
}

function skeleton () {
    local PROBLEM=$1
    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/skeleton ${PROBLEM} --avisfukudaformat > ${PROBLEM}.skeleton.log 2> ${PROBLEM}.skeleton.err
    RET_VALUE=$?

    mv ${PROBLEM}.out ${PROBLEM}.ref
    mv ${PROBLEM}.skeleton.* ${DOC}
}

function cddr () {
    local PROBLEM=$1
    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT  ${BIN}/cddr+ ${PROBLEM} > ${PROBLEM}.cddr.log 2> ${PROBLEM}.cddr.err
    RET_VALUE=$?

    mv ${PROBLEM}.cddr.* ${DOC}
}

function ppl () {
    local PROBLEM=$1
    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/ppl_lcdd ${PROBLEM} > ${PROBLEM}.ppl.log 2> ${PROBLEM}.ppl.err
    RET_VALUE=$?

    mv ${PROBLEM}.ppl.* ${DOC}
}

function polco () {
    local PROBLEM=$1
    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT java -jar ${BIN}/polco.jar -arithmetic bigint bigint bigint -maxthreads 1 -kind cdd -in ${PROBLEM} -out text ${PROBLEM}.out > ${PROBLEM}.polco.log 2> ${PROBLEM}.polco.err
    RET_VALUE=$?

    mv ${PROBLEM}.polco.* ${DOC}
}

function porta () {
    local PROBLEM=$1
    cp ${PROBLEM%.*}.ieq current.ieq

    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/xporta -Tl current.ieq > ${PROBLEM}.porta.log 2> ${PROBLEM}.porta.err
    RET_VALUE=$?

    mv ${PROBLEM}.porta.* ${DOC}
}

function four-ti-two () {
    local PROBLEM=$1

    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/rays -p arb ${PROBLEM%.*} 2> ${PROBLEM}.four-ti-two.err > ${PROBLEM}.four-ti-two.log
    RET_VALUE=$?

    mv ${PROBLEM}.four-ti-two.* ${DOC}
}

function addibit-bp-trees () {
    local PROBLEM=$1

    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/addibit ${PROBLEM} settings_bp-trees.conf > ${PROBLEM}.addibit-bp-trees.log 2> ${PROBLEM}.addibit-bp-trees.err
    RET_VALUE=$?

    mv ${PROBLEM}.addibit-comb.* ${DOC}
}

function addibit-bp-qbn-trees () {
    local PROBLEM=$1

    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/addibit ${PROBLEM} settings_bp-qbn-trees.conf > ${PROBLEM}.addibit-bp-qbn-trees.log 2> ${PROBLEM}.addibit-bp-qbn-trees.err
    RET_VALUE=$?

    mv ${PROBLEM}.addibit-comb.* ${DOC}
}

function addibit-vp-trees () {
    local PROBLEM=$1

    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/addibit ${PROBLEM} settings_vp-trees.conf > ${PROBLEM}.addibit-vp-trees.log 2> ${PROBLEM}.addibit-vp-trees.err
    RET_VALUE=$?

    mv ${PROBLEM}.addibit-comb.* ${DOC}
}

function addibit-pst () {
    local PROBLEM=$1

    /usr/bin/time -f "%M" -o mem_buffer timeout $TIMEOUT ${BIN}/addibit ${PROBLEM} settings_bp-qbn-trees_pst.conf > ${PROBLEM}.addibit-pst.log 2> ${PROBLEM}.addibit-pst.err
    RET_VALUE=$?

    mv ${PROBLEM}.addibit-pst.* ${DOC}
}

function addibit-alg () {
    local PROBLEM=$1

    /usr/bin/time -f "%M" -o mem_buffer timeout $2 ${BIN}/addibit ${PROBLEM} settings_bp-qbn-trees_alg.conf > ${PROBLEM}.addibit-alg.log 2> ${PROBLEM}.addibit-alg.err
    RET_VALUE=$? 

    mv ${PROBLEM}.addibit-alg.* ${DOC}
}

function print_multiple () {
 str=$1
 num=$2
 v=$(printf "%-${num}s" "$str")
 echo -n "${v// /$str}"
}

function print_caption () {
    local CAPTION=$1
    local first_run=$(( (80-${#CAPTION}) / 2 ))
    local second_run=$(( $first_run + ( 80 - ${#CAPTION}) % 2 ))

    echo -n "+"
    print_multiple "-" $first_run
    print_multiple "-" ${#CAPTION}
    print_multiple "-" $second_run
    echo "+"

    echo -n "|"
    print_multiple " " $first_run
    echo -n ${CAPTION}
    print_multiple " " $second_run
    echo "|"

    echo -n "+"
    print_multiple "-" $first_run
    print_multiple "-" ${#CAPTION}
    print_multiple "-" $second_run
    echo "+"    
}

function run_single () {
    local PROGRAM=$1
    local COLUMN=$2
    local PROBLEM=$3
    local DESCRIPTION=$4

    local LOG="${DOC}/${DESCRIPTION}.ine.${PROGRAM}.log"
    local LOG_PATH=`basename ${LOG}`

    if [[ ${OVERALL_TIME[${COLUMN}]} != "-" ]]; then
        local TIMEOUT=$( echo "$TIME_LIMIT - ${OVERALL_TIME[${COLUMN}]}" | bc )

        echo -n "|  ${PROGRAM} "
        print_multiple "." $(( 59 - ${#PROGRAM} ))

        local START=$(date +%s.%N)
        ${PROGRAM} ${PROBLEM} ${TIMEOUT}

        if [[ ${RET_VALUE} == "124" ]]; then
                echo -e ".....${YELLOW_COLOR}  Timeout${RESET_COLOR}    |"     
                TABLE="${TABLE}<td><a href=\"${LOG_PATH}\">TIMEOUT</a></td>"
                OVERALL_TIME[${i}]="-"  
                echo -n "- " >> ${MEM_PLOT}
        else 
            if [[ ${RET_VALUE} != "0" ]]; then
                echo -e ".......${RED_COLOR}  Error${RESET_COLOR}    |"
                TABLE="${TABLE}<td><a href=\"${LOG_PATH}\">Error &#9889</a></td>"

                OVERALL_TIME[${i}]="-"

                echo -n "- " >> ${MEM_PLOT}
            else
                cat mem_buffer | tr '\n' ' ' >> ${MEM_PLOT}

                local END=$(date +%s.%N)
                local TIME=$( echo "scale=3;(${END} - ${START})/1" | bc )

                print_multiple "." $(( 9 - ${#TIME} ))
                echo -e "${GREEN_COLOR} ${TIME} sec${RESET_COLOR}    |"

                OVERALL_TIME[${COLUMN}]=$(echo "scale=3;(${OVERALL_TIME[${COLUMN}]} + ${TIME})/1" | bc)

                TABLE="${TABLE}<td><a href=\"${LOG_PATH}\">${TIME}</a></td>"
            fi
        fi
    else
        TABLE="${TABLE}<td>-</td>"
        echo -n "- " >> ${MEM_PLOT}
    fi
}

function run () {
    local PROBLEM=$1
    local DESCRIPTION=`basename ${PROBLEM} .ine`

    ${BIN}/addibit-export 4ti2 ${PROBLEM}
    ${BIN}/addibit-export porta ${PROBLEM}

    TABLE="${TABLE}<td>${DESCRIPTION}</td>"
    OUTPUT_FILE_SIZE=0

    local TIME_TABLE=""

    print_caption ${DESCRIPTION}

    echo -n "${DESCRIPTION} ${PROBLEM_COUNTER} " >> ${MEM_PLOT}

    for (( i = 0; i < ${#REFERENCE_PROGRAM[@]}; i++ )); do
        run_single ${REFERENCE_PROGRAM[${i}]} ${i} ${PROBLEM} ${DESCRIPTION} 2>> /dev/null
    done

    for (( i = 0; i < ${#CONFIGURATION[@]}; i++ )); do
        STRATEGY=${CONFIGURATION[${i}]}

        local LOG="${DOC}/${DESCRIPTION}.${STRATEGY}.log"

        echo -n "|  addibit with ${STRATEGY} "
        print_multiple "." $(( 46 - ${#STRATEGY} ))

        local START=$(date +%s.%N)
        ${BIN}/addibit ${PROBLEM} settings_${STRATEGY}.conf > ${LOG}
        local END=$(date +%s.%N)
        local TIME=$( echo "scale=3;(${END} - ${START})/1" | bc )

        cp gnuplot.dat ${PLOT_FOLDER}/${DESCRIPTION}.${STRATEGY}.dat

        local LOG_PATH=`basename ${LOG}`
        TABLE="${TABLE}<td><a href=\"${LOG_PATH}\">${TIME}</a></td>"

        OVERALL_TIME[(${i}+${#REFERENCE_PROGRAM[@]})]=$(echo "scale=3;(${OVERALL_TIME[(${i}+${#REFERENCE_PROGRAM[@]})]} + ${TIME})/1" | bc)

        if [ ! -z ${REFERENCE_PROGRAM} ]; then
            REF_DIFF=`diff -b <(sed 1d ${PROBLEM}.out | sort -n) <(sed 1,4d ${PROBLEM}.ref | sort -n)`
            if [ "${REF_DIFF}" != "" ]; then
                echo -e "${RED_COLOR} inconsistent output differs from reference!${RESET_COLOR}"
            fi
        fi

        if (( ${OUTPUT_FILE_SIZE} > 0 )); then
            if (( ${OUTPUT_FILE_SIZE} != `stat -c%s ${PROBLEM}.out` )); then
                echo -e "${RED_COLOR} inconsistent output file!${RESET_COLOR}"
            else
                print_multiple "." $(( 9 - ${#TIME} ))
                echo -e "${GREEN_COLOR} ${TIME} sec${RESET_COLOR}    |"
            fi
        else
            OUTPUT_FILE_SIZE=`stat -c%s ${PROBLEM}.out`
            print_multiple "." $(( 9 - ${#TIME} ))
            echo -e "${GREEN_COLOR} ${TIME} sec${RESET_COLOR}    |"
        fi
    done  

    echo "${DESCRIPTION} ${PROBLEM_COUNTER} ${OVERALL_TIME[@]}" >> ${TIME_PLOT}
    echo " " >> ${MEM_PLOT}
    PROBLEM_COUNTER=$((${PROBLEM_COUNTER}+1))
    
    echo -n "+"
    print_multiple "-" 80
    echo "+"
}

TABLE="<head><style type=\"text/css\"><!--@import url(\"style.css\");--></style></head>\n"
TABLE="${TABLE}<body>\n<table id=\"results_table\">\n"
# Table head
TABLE="${TABLE}\t<thead>\n"
TABLE="${TABLE}\t\t<th></th>\n"

echo -n "#" >> ${MEM_PLOT}
echo -n "#" >> ${TIME_PLOT}
for (( i = 0; i < ${#REFERENCE_PROGRAM[@]}; i++ )); do    
    TABLE="${TABLE}\t\t<th>${REFERENCE_PROGRAM[${i}]}</th>\n"
    OVERALL_TIME[${i}]=0

    echo -n " ${REFERENCE_PROGRAM[${i}]}" >> ${MEM_PLOT}
    echo -n " ${REFERENCE_PROGRAM[${i}]}" >> ${TIME_PLOT}
done 
for (( i = 0; i < ${#CONFIGURATION[@]}; i++ )); do
    STRATEGY=${CONFIGURATION[${i}]}
    
    TABLE="${TABLE}\t\t<th>${STRATEGY}</th>\n"
    OVERALL_TIME[(${i}+${#REFERENCE_PROGRAM[@]})]=0
done 
TABLE="${TABLE}\t</thead>\n"

echo "" >> ${MEM_PLOT}
echo "" >> ${TIME_PLOT}

traverse $ROOT

TABLE="${TABLE}\t\t<tr><td>Total:</td>"
for (( i = 0; i < ${#REFERENCE_PROGRAM[@]}; i++ )); do    
    TABLE="${TABLE}\t\t<td>${OVERALL_TIME[${i}]}</td>\n"
done 
for (( i = 0; i < ${#CONFIGURATION[@]}; i++ )); do
    TABLE="${TABLE}\t\t<td>${OVERALL_TIME[(${i}+${#REFERENCE_PROGRAM[@]})]}</td>\n"
done 
TABLE="${TABLE}</tr>\n</table></body>"
echo -e ${TABLE} > "${DOC}/benchmark.html"
