/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_BIT_LIST__
#define __DD_BIT_LIST__

#include <vector>
#include <string>
#include <string.h>

#include "DDTweaks.h"
#include "DDLogging.h"

#include "DDBitCountTable.h"

#if defined _LP64 || defined __LP64__
#define __SINGLE_BIT__ ( static_cast<unsigned long>( 0x0000000000000001 ))
#define __BLOCK_SIZE__ 64
#else 
#define __SINGLE_BIT__ 0x00000001
#define __BLOCK_SIZE__ 32
#endif // defined _LP64 || defined __LP64__

#ifdef __DOUBLE_BLOCK_OPT__
#define __DEFAULT_BLOCK_COUNT__ 2
#else
#define __DEFAULT_BLOCK_COUNT__ 1
#endif // __DOUBLE_BLOCK_OPT__

namespace double_description {

	class BitList {

		private:
		unsigned long mBlocks;
		
#ifdef __DOUBLE_BLOCK_OPT__
		unsigned long mFirstBlock;
		unsigned long mSecondBlock;
#else
		
	#ifdef __SINGLE_BLOCK_OPT__
		unsigned long mFirstBlock;
	#endif // __SINGLE_BLOCK_OPT__

#endif // __DOUBLE_BLOCK_OPT__

		unsigned long* data;  

		public:

#ifdef __SINGLE_BLOCK_OPT__

		BitList () :	
			mBlocks( __DEFAULT_BLOCK_COUNT__ ),
			mFirstBlock( 0 ) {}

		BitList ( const unsigned long blocks ) :	
			mBlocks( blocks ),
			mFirstBlock( 0 ) {

			if ( __builtin_expect( mBlocks > 1, 0 ) ) {
				data = new unsigned long[ mBlocks - 1 ];
			
				for ( unsigned short block = 0; block < mBlocks - 1; block++ ) {
					data[block] = 0;
				}	 
			}
		}

		BitList ( const unsigned long blocks, const unsigned long value ) :	
			mBlocks( blocks ),
			mFirstBlock( value ) {

			if ( __builtin_expect( mBlocks > 1, 0 ) ) {
				data = new unsigned long[ mBlocks - 1 ];
			
				for ( unsigned short block = 0; block < mBlocks - 1; block++ ) {
					data[block] = value;
				}	 
			}
		}

        BitList ( const std::vector<bool> & vec ) :
            mBlocks( getBlocks( vec.size() ) ),
            mFirstBlock( 0 ) {
                
            init( mBlocks, 0 );

            for ( unsigned int index = 0; index < vec.size(); index++ ) {
                if ( vec[ index ] ) {
                    setBit( index );
                }
            }
        }

		BitList ( const BitList& other ) :	
			mBlocks( other.mBlocks ),
			mFirstBlock( other.mFirstBlock ) {

			if ( __builtin_expect( mBlocks > 1, 0 ) ) {
				data = new unsigned long[ mBlocks - 1 ];
			
				for ( unsigned short block = 0; block < mBlocks - 1; block++ ) {
					data[block] = other.data[block];
				}	 
			}
		}
#else

	#ifdef __DOUBLE_BLOCK_OPT__

		BitList () :	
			mBlocks( __DEFAULT_BLOCK_COUNT__ ),
			mFirstBlock( 0 ),
			mSecondBlock( 0 ) {}

		BitList ( const unsigned long blocks ) :	
			mBlocks( blocks ),
			mFirstBlock( 0 ),
			mSecondBlock( 0 ) {

			if ( __builtin_expect( blocks > 2, 0 ) ) {
				data = new unsigned long[ blocks - 2 ];
			
				for ( unsigned short block = 0; block < mBlocks - 2; block++ ) {
					data[block] = 0;
				}	 
			}
		}

		BitList ( const unsigned long blocks, const unsigned long value ) :	
			mBlocks( blocks ),
			mFirstBlock( value ),
			mSecondBlock( value ) {

			if ( __builtin_expect( blocks > 2, 0 ) ) {
				data = new unsigned long[ blocks - 2 ];
			
				for ( unsigned short block = 0; block < mBlocks - 2; block++ ) {
					data[block] = value;
				}	 
			}
		}

        BitList ( const std::vector<bool> & vec ) :
            mBlocks( getBlocks( vec.size() ) ),
            mFirstBlock( 0 ),
            mSecondBlock( 0 ) {

            init( mBlocks, 0 );

            for ( unsigned int index = 0; index < vec.size(); index++ ) {
                if ( vec[ index ] ) {
                    setBit( index );
                }
            }
        }

		BitList ( const BitList& other ) :	
			mBlocks( other.mBlocks ),
			mFirstBlock( other.mFirstBlock ),
			mSecondBlock( other.mSecondBlock ) {

			if ( __builtin_expect( mBlocks > 2, 0 ) ) {
				data = new unsigned long[ mBlocks - 2 ];
			
				for ( unsigned short block = 0; block < mBlocks - 2; block++ ) {
					data[block] = other.data[block];
				}	 
			}
		}
	#else

		BitList () :	
			mBlocks( __DEFAULT_BLOCK_COUNT__ ) {
			
			data = new unsigned long[ __DEFAULT_BLOCK_COUNT__ ];	
			data[0] = 0;

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] = 0;
			} 
		}


		BitList ( const unsigned long blocks ) :	
			mBlocks( blocks ) {

			data = new unsigned long[blocks];
			data[0] = 0;

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] = 0;
			} 
		}

		BitList ( const unsigned long blocks, const unsigned long value ) :	
			mBlocks( blocks ) {

			data = new unsigned long[blocks];
			data[0] = value;

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] = value;
			} 
		}


		BitList ( const BitList& other ) :	
			mBlocks( other.mBlocks ) {

			data = new unsigned long[mBlocks];
			data[0] = other.data[0];

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] = other.data[block];
			} 
		}

	#endif //__DOUBLE_BLOCK_OPT__

#endif //__SINGLE_BLOCK_OPT__

		~BitList () {
#ifdef __DOUBLE_BLOCK_OPT__
				if ( unlikely ( mBlocks > 2 ) ) {
					delete[] data;
				}
#else
	#ifdef __SINGLE_BLOCK_OPT__
				if ( unlikely ( mBlocks > 1 ) ) {
					delete[] data;
				}
	#else
			delete[] data;
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__
		}

		static const BitCountTable BIT_COUNT_TABLE;

		static unsigned long getBlocks ( const unsigned long capacity ) {
			unsigned long block_count = capacity / __BLOCK_SIZE__ + (( capacity % __BLOCK_SIZE__ == 0 ) ? 0 : 1 );

#ifdef __DOUBLE_BLOCK_OPT__
			return ( block_count > 2 ) ? block_count : 2; 
#else
			return block_count;
#endif // __DOUBLE_BLOCK_OPT__
		}

		unsigned long getBlocks () const {
			return mBlocks;
		}

		unsigned long getCapacity () const {
			return mBlocks * __BLOCK_SIZE__;
		}

		void init ( const unsigned short blocks, const unsigned long data );

		void resize ( const unsigned long blocks );

		bool isEmpty () const {
			bool result = ( getData( 0 ) == 0 );

			for ( unsigned long block = 1; block < mBlocks && result; block++ ) {
				if ( getData( block ) != 0 ) {
					return false;
				}
			}

			return result;
		}

		bool isFull () const {
			bool result = ( getData( 0 ) == ULONG_MAX );

			for ( unsigned long block = 1; block < mBlocks && result; block++ ) {
				if ( getData( block ) != ULONG_MAX ) {
					result = false;
					break;
				}
			}

			return result;
		}

		bool isUnionFull ( const BitList & other ) const {
			bool result = ( (getData( 0 ) | other.getData( 0 )) == ULONG_MAX );

			for ( unsigned long block = 1; block < mBlocks && result; block++ ) {
				if ( (getData( block ) | other.getData( block )) != ULONG_MAX ) {
					result = false;
					break;
				}
			}

			return result;
		}

		 const unsigned long & getData ( const unsigned long block ) const {

#ifdef __DOUBLE_BLOCK_OPT__

			if ( __builtin_expect ( block > 1, 0 ) ) {
				return data[block-2];
			} else if ( block == 0 ) {
				return mFirstBlock;
			} else {
				return mSecondBlock;
			}

#else 

	#ifdef  __SINGLE_BLOCK_OPT__

			if ( __builtin_expect ( block > 0, 0 ) ) {
				return data[block-1];
			} else {
				return mFirstBlock;
			}

	#else
	
			return data[block];

	#endif //__SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

		}

		 unsigned long & getData ( const unsigned long block ) {

#ifdef __DOUBLE_BLOCK_OPT__

			if ( __builtin_expect ( block > 1, 0 ) ) {
				return data[block-2];
			} else if ( block == 0 ) {
				return mFirstBlock;
			} else {
				return mSecondBlock;
			}

#else 

	#ifdef  __SINGLE_BLOCK_OPT__

			if ( __builtin_expect ( block > 0, 0 ) ) {
				return data[block-1];
			} else {
				return mFirstBlock;
			}

	#else
	
			return data[block];

	#endif //__SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__
		}

		const BitList operator& ( const BitList& rhs ) const {
			BitList result( mBlocks );

			for ( unsigned long i = 0; i < mBlocks; i++ ) {
				result[i] = getData( i ) & rhs[i];
			}

			return result;
		}

		const BitList operator| ( const BitList& rhs ) const {
			BitList result( mBlocks );

			for ( unsigned long i = 0; i < mBlocks; i++ ) {
				result[i] = getData( i ) | rhs[i];
			}

			return result;
		}

		const BitList operator^ ( const BitList& rhs ) const {
			BitList result( mBlocks );

			for ( unsigned long i = 0; i < mBlocks; i++ ) {
				result[i] = getData( i ) ^ rhs[i];
			}

			return result;
		}

		const BitList operator~ () const {
			BitList result( mBlocks );

			for ( unsigned long i = 0; i < mBlocks; i++ ) {
				result[i] = ~getData( i );
			}

			return result;
		}

        const BitList operator- ( const BitList & rhs ) const {
            return (*this) ^ ( (*this) & rhs );
        }

		bool operator== ( const BitList & rhs ) const {

			for ( unsigned long i = 0; i < mBlocks; i++ ) {
				if ( getData( i ) != rhs[i] ) {
					return false;
			    }
            } 

			return true;
		}

		bool operator!= ( const BitList & rhs ) const {
	
			return !operator==( rhs );
		}

        bool lex_less ( const BitList & rhs ) const {
            unsigned int block = 0;                  
      
			for ( ; block < mBlocks - 1 && getData( block ) == rhs.getData( block ); block++ );

            return getData( block ) < rhs.getData( block );
        }

		void operator &= ( const BitList& rhs ) {

            if ( __builtin_expect( mBlocks != rhs.mBlocks, false ) ) {
                resize( rhs.mBlocks );
            }

#ifdef __DOUBLE_BLOCK_OPT__
			mFirstBlock &= rhs.mFirstBlock;
			mSecondBlock &= rhs.mSecondBlock; 

			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
				data[block-2] &= rhs[block];
#else
	#ifdef __SINGLE_BLOCK_OPT__
			mFirstBlock &= rhs.mFirstBlock;

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
				data[block-1] &= rhs[block];
	#else
			data[0] &= rhs[0];			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] &= rhs[block];
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__
			}
		}

		void operator |= ( const BitList & rhs ) {

            if ( __builtin_expect( mBlocks != rhs.mBlocks, false ) ) {
                resize( rhs.mBlocks );
            }

#ifdef __DOUBLE_BLOCK_OPT__
			mFirstBlock |= rhs.mFirstBlock;
			mSecondBlock |= rhs.mSecondBlock; 

			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), false ); block++ ) {
				data[block-2] |= rhs[block];
#else
	#ifdef __SINGLE_BLOCK_OPT__
			mFirstBlock |= rhs.mFirstBlock;

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), false ); block++ ) {
				data[block-1] |= rhs[block];
	#else
			data[0] |= rhs[0];			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] |= rhs[block];
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__
			}
		}

		void operator= ( const BitList & rhs ) {

            if ( __builtin_expect( mBlocks != rhs.mBlocks, false ) ) {
                resize( rhs.mBlocks );
            }

#ifdef __DOUBLE_BLOCK_OPT__
			mFirstBlock = rhs.mFirstBlock;
			mSecondBlock = rhs.mSecondBlock; 

			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
			    data[block-2] = rhs[block];
            }
#else
	#ifdef __SINGLE_BLOCK_OPT__
			mFirstBlock = rhs.mFirstBlock;

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
			   data[block-1] = rhs[block];
            }
	#else
			data[0] = rhs[0];			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] = rhs[block];
            }
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__
		}

		const unsigned long& operator[] ( const unsigned long index ) const {
			return getData( index );
		}
		unsigned long& operator[] ( const unsigned long index ) {
			return getData( index );
		}

		void clear () {

#ifdef __DOUBLE_BLOCK_OPT__
			mFirstBlock = 0;
			mSecondBlock = 0; 

			for ( unsigned short block = 2; __builtin_expect( block < mBlocks, false ); block++ ) {
				data[block-2] = 0;
#else
	#ifdef __SINGLE_BLOCK_OPT__
			mFirstBlock = 0;

			for ( unsigned short block = 1; __builtin_expect( block < mBlocks, false ); block++ ) {
				data[block-1] = 0;
	#else
			data[0] = 0;			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
				data[block] = 0;
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__
			}
		}

		unsigned short getPopulationCount () const {
			unsigned long count;

#ifdef __DOUBLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock ) + BitList::BIT_COUNT_TABLE.lookUp ( mSecondBlock );

			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
#else
	#ifdef __SINGLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock );

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
	#else
			count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] );			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) );
			}

			return count;
		}

		unsigned short getInvPopulationCount () const {
			unsigned long count;

#ifdef __DOUBLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( ~mFirstBlock ) + BitList::BIT_COUNT_TABLE.lookUp ( ~mSecondBlock );

			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
#else
	#ifdef __SINGLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( ~mFirstBlock );

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
	#else
			count = BitList::BIT_COUNT_TABLE.lookUp ( ~data[0] );			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

				count += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) );
			}

			return count;
		}

        unsigned short getHammingDistance ( const BitList & other ) const {
			unsigned long count;

#ifdef __DOUBLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock xor other.mFirstBlock ) +
					BitList::BIT_COUNT_TABLE.lookUp ( mSecondBlock xor other.mSecondBlock );

			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
#else
	#ifdef __SINGLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock xor other.mFirstBlock );

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
	#else
			count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] xor other.data[0] );			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) xor other.getData( block ) );
			}

			return count;
        }

		unsigned short getCommonPopCnt ( const BitList & other ) const {
			unsigned long count;

#ifdef __DOUBLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock & other.mFirstBlock ) +
					BitList::BIT_COUNT_TABLE.lookUp ( mSecondBlock & other.mSecondBlock );

			for ( unsigned short block = 2; unlikely ( block < mBlocks ); block++ ) {
#else
	#ifdef __SINGLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock & other.mFirstBlock );

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
	#else
			count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] & other.data[0] );			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) & other.getData( block ) );
			}

			return count;
		}

		unsigned short getCommonPopCnt ( const BitList & other, unsigned short & inv_pop_cnt ) const {
			unsigned long count;

#ifdef __DOUBLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock & other.mFirstBlock ) +
					BitList::BIT_COUNT_TABLE.lookUp ( mSecondBlock & other.mSecondBlock );
            inv_pop_cnt = BitList::BIT_COUNT_TABLE.lookUp ( ~mFirstBlock & other.mFirstBlock ) +
					      BitList::BIT_COUNT_TABLE.lookUp ( ~mSecondBlock & other.mSecondBlock );

			for ( unsigned short block = 2; unlikely ( block < mBlocks ); block++ ) {
#else
	#ifdef __SINGLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( mFirstBlock & other.mFirstBlock );
            inv_pop_cnt = BitList::BIT_COUNT_TABLE.lookUp ( ~mFirstBlock & other.mFirstBlock );

			for ( unsigned short block = 1; unlikely ( block < mBlocks ); block++ ) {
	#else
			count = BitList::BIT_COUNT_TABLE.lookUp ( data[0] & other.data[0] );	
            inv_pop_cnt = BitList::BIT_COUNT_TABLE.lookUp ( ~data[0] & other.data[0] );		

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

				count += BitList::BIT_COUNT_TABLE.lookUp ( getData( block ) & other.getData( block ) );
                inv_pop_cnt += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) & other.getData( block ) );
			}

			return count;
		}

		unsigned short getInvCommonPopCnt ( const BitList & other ) const {
			unsigned long count;

#ifdef __DOUBLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( ~mFirstBlock & other.mFirstBlock ) +
					BitList::BIT_COUNT_TABLE.lookUp ( ~mSecondBlock & other.mSecondBlock );

			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
#else
	#ifdef __SINGLE_BLOCK_OPT__
			count = BitList::BIT_COUNT_TABLE.lookUp ( ~mFirstBlock & other.mFirstBlock );

			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
	#else
			count = BitList::BIT_COUNT_TABLE.lookUp ( ~data[0] & other.data[0] );			

			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

				count += BitList::BIT_COUNT_TABLE.lookUp ( ~getData( block ) & other.getData( block ) );
			}

			return count;
		}

		unsigned short getFirstSet () const {
			unsigned long block = 0;

			for ( block = 0; __builtin_expect( block < mBlocks-1 && !getData(block), 0 ); block++ );

			return block * __BLOCK_SIZE__ + __builtin_ctzl( getData( block ));
		}

		unsigned short getFirstUnset () const {
			unsigned long block = 0;

			for ( block = 0; unlikely( block < mBlocks-1 && !(~getData( block ))); block++ );

			return block * __BLOCK_SIZE__ + __builtin_ctzl( ~getData( block ));
		}

        unsigned short getNextSet ( unsigned short start ) const {
            unsigned long block = start / __BLOCK_SIZE__;
            unsigned long offset = start % __BLOCK_SIZE__;
            unsigned short result;

            register unsigned long pattern = ( ULONG_MAX << offset ) & getData( block );

            if ( __builtin_expect( pattern != 0, 1 ) ) {
                result = block * __BLOCK_SIZE__ + __builtin_ctzl( pattern );
            } else if ( __builtin_expect( block + 1 >= mBlocks, 0 ) ) {
                result = getCapacity();
            } else {
                for ( block++ ; __builtin_expect( block < mBlocks - 1 && getData( block ) != 0, 0 ); block++ );

                if ( getData( block ) != 0 ) {
                    result = block * __BLOCK_SIZE__ + __builtin_ctzl( getData( block ));
                } else {
                    result = getCapacity();
                }
            }

            return result;
        }

		bool isSet ( const unsigned long position ) const {
			return ( getData( position / __BLOCK_SIZE__ ) & ( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) ) );
		}

		void setBit ( const unsigned long position ) {
			getData( position / __BLOCK_SIZE__ ) |= ( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) );
		}

		void opOR ( const unsigned long block, const unsigned long mask ) {
			getData( block ) |= mask;
		}

		void setBits ( const unsigned long begin, const unsigned long end ) {

			for ( unsigned long index = begin; index < end; index++ ) {
				setBit( index );
			}
		}

		void unsetBits ( const unsigned long begin, const unsigned long end ) {
			for ( unsigned long index = begin; index < end; index++ ) {
				unsetBit( index );
			}			
		}

		void unsetBitsFrom ( const unsigned long begin ) {
			for ( unsigned long index = begin; index < getCapacity() ; index++ ) {
				unsetBit( index );
			}			
		}

		void unsetBit ( const unsigned long position ) {	
			getData( position/ __BLOCK_SIZE__ ) &= ~( __SINGLE_BIT__ << ( position % __BLOCK_SIZE__ ) );
		}

		void opXOR ( const unsigned long block, const unsigned long mask ) {
			getData( block ) ^= mask;
		}

		bool containsUnexpected ( const unsigned long * const map ) const {

#ifdef __DOUBLE_BLOCK_OPT__
			bool result = __builtin_expect( (( map[0] & mFirstBlock ) == map[0] ), 0 ) && __builtin_expect( (( map[1] & mSecondBlock ) == map[1] ), 0 );
#else 
	#ifdef __SINGLE_BLOCK_OPT__
			bool result = __builtin_expect( (( map[0] & mFirstBlock ) == map[0] ), 0 );
	#else
			bool result = ( map[0] & data[0] ) == map[0];
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

	
#ifdef __DOUBLE_BLOCK_OPT__
			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ) && __builtin_expect( result, 0 ); block++ ) {
#else

	#ifdef __SINGLE_BLOCK_OPT__
			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ) && __builtin_expect( result, 0 ); block++ ) {
	#else
			for ( unsigned short block = 1;  block < mBlocks && __builtin_expect( result, 0 ); block++ ) {
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

				result = ( map[block] & getData( block ) ) == map[block];		
			} 

			return result;
		}

		bool contains ( const BitList & map ) const {

#ifdef __DOUBLE_BLOCK_OPT__
			bool result = (( map[0] & mFirstBlock ) == map[0] ) && (( map[1] & mSecondBlock ) == map[1] );
#else 
	#ifdef __SINGLE_BLOCK_OPT__
			bool result = ( map[0] & mFirstBlock ) == map[0];
	#else
			bool result = ( map[0] & data[0] ) == map[0];
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__

	
#ifdef __DOUBLE_BLOCK_OPT__
			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ) && result; block++ ) {
				result = ( map[block] & getData( block ) ) == map[block];	
#else

	#ifdef __SINGLE_BLOCK_OPT__
			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ) && result; block++ ) {
				result = ( map[block] & getData( block ) ) == map[block];	
	#else
			for ( unsigned short block = 1;  block < mBlocks && result; block++ ) {
				result = ( map[block] & getData( block ) ) == map[block];
	#endif // __SINGLE_BLOCK_OPT__
#endif //__DOUBLE_BLOCK_OPT__	
			} 

			return result;
		}

		const std::string toString () const;
        const std::string toString ( const unsigned short length ) const;
	};
}

#endif // __DD_BIT_LIST__
