/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_BIT_PATTERN_TREE_CACHE_ENTRY__
#define __DD_BIT_PATTERN_TREE_CACHE_ENTRY__

#include <vector>
#include <sstream>
#include <algorithm>
#include <math.h>

#include "DDStats.h"

#include "DDOptions.h"

#include "DDBitList.h"
#include "DDBitCountTable.h"
#include "DDBitSelector.h"

#include "DDRayReferenceContainer.h"
#include "DDRayRefReleaseVisitor.h"

namespace double_description {
	
	template <class RAY_TYPE>
	class BitPatternTreeCacheEntry {

		private:
        const unsigned short mRows;
		unsigned short mPlane;
		unsigned short mBlocks;
        unsigned int mMinCommonBits;

		RayReferenceContainer mPositiveDegEntries;
        RayReferenceContainer mPositiveNonDegEntries;
		RayReferenceContainer mZeroEntries;
        RayReferenceContainer mNegativeEntries;

        BitList mCutMap;

		public:
		BitPatternTreeCacheEntry ( const unsigned short rows,
                                   const unsigned short columns,
                                   const unsigned short current ) :
            mRows( rows ),
			mPlane( current ),
			mBlocks( BitList::getBlocks( rows ) ),
            mMinCommonBits( columns - 2 ),
            mCutMap( mBlocks, 0 ) {

                mCutMap.setBits( 0, mPlane );
            }

		BitPatternTreeCacheEntry ( const BitPatternTreeCacheEntry & other ) :
            mRows( other.mRows ),
			mPlane( other.mPlane ),
			mBlocks( other.mBlocks ),
            mMinCommonBits( other.mMinCommonBits ),
            mCutMap( other.mCutMap ) {}

        ~BitPatternTreeCacheEntry () {
            const RayRefReleaseVisitor<RAY_TYPE> release_visitor;
            mNegativeEntries.modify( release_visitor );
        }

        const RayReferenceContainer & positive_nondeg () const {
            return mPositiveNonDegEntries;
        }

        const RayReferenceContainer & positive_deg () const {
            return mPositiveDegEntries;
        }

        const RayReferenceContainer & negative () const {
            return mNegativeEntries;
        }

        const RayReferenceContainer & degenerate () const {
            return mZeroEntries;
        }

		void insertPositive ( const RAY_TYPE* const ray ) {
            const RayReference ref( static_cast<const void*>( ray ), ray->getActiveSet() & mCutMap );
            
            if ( ref.activeSet.getPopulationCount() == mMinCommonBits + 1 ) {
                mPositiveNonDegEntries.insert( ref );
            } else {    
			    mPositiveDegEntries.insert( ref );
                mZeroEntries.insert( ref );
            }
		}

		void insertNegative ( const RAY_TYPE* const ray ) {
            const RayReference ref( static_cast<const void*>( ray ), ray->getActiveSet() & mCutMap );

			mNegativeEntries.insert( ref );

            if ( ref.activeSet.getPopulationCount() > mMinCommonBits + 1 ) {
                mZeroEntries.insert( ref );
            }
		}

		void insertZero ( const RAY_TYPE* const ray ) {
            const RayReference ref( static_cast<const void*>( ray ), ray->getActiveSet() & mCutMap );
	
			mZeroEntries.insert( ref );
		}

		unsigned int getPositiveCount () const {
			return mPositiveDegEntries.size() + mPositiveNonDegEntries.size();
		}

		unsigned int getNegativeCount () const {
			return mNegativeEntries.size();
		}
	};
}

#endif // __DD_BIT_PATTERN_TREE_CACHE_ENTRY__
