/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_CALCULATION_MATRIX__
#define __DD_CALCULATION_MATRIX__

#include "DDBitList.h"
#include "DDMatrix.h"
#include "DDUtilsFactory.h"

#include "DDLogging.h"

namespace double_description {

	class CalculationMatrix : protected Matrix {
		
		private:
		unsigned int mRowLimit;
		unsigned int mColumnLimit;

		protected:
		CalculationMatrix ( const unsigned int rows, const unsigned int columns );
		virtual ~CalculationMatrix () {}

		void transformToUpperTriangularMatrix () {

			for ( unsigned long row = 0, column = 0; ( row < getRows() - 1 ) && ( column < getColumns() - 1 ); column++ ) {

				if ( findPivot( row, column ) ) {
                    eliminateColumnDown( row++, getRows(), column );
				}
			}
		}

        void simplifyUpperTriangularMatrix ( const unsigned short end ) {
            const unsigned int columns = getColumns();

            for ( unsigned int r_index = 0; r_index < end; r_index++ ) {
                getRow( r_index ).simplify( r_index, columns );
            }
        }

		void transformToDiagonalMatrix () {

            for ( int row = getRows(); row > 0; row-- ) {
                if ( isNonZeroAt( row - 1, row - 1 ) ) {
                    eliminateColumnUp( 0, row, row - 1 );
                }
            }
		}

		void prepareRectangleForElimination ( const unsigned long beginRow, const unsigned long column ) {
			prepareRectangleForElimination( beginRow, getRows(), column );
		}

		void prepareRectangleForElimination ( const unsigned long beginRow, const unsigned long endRow, const unsigned long column ) {

			UtilsFactory::CALC_UTILS.startIncrementalLCM ( getValue( beginRow, column ) );

			for ( unsigned long i = beginRow + 1; i < endRow; i++ ) {
				UtilsFactory::CALC_UTILS.calcIncrementalLCM ( getValue ( i, column ) );
			}
			UtilsFactory::CALC_UTILS.convertResultToNegativeValue();
			multiplyRectangle( beginRow, endRow, column, UtilsFactory::CALC_UTILS.getResult() );
		}

		void prepareTriangleForElimination ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn ) {

			UtilsFactory::CALC_UTILS.startIncrementalLCM ( getValue( beginRow, pivotColumn ) );

			for ( unsigned long i = beginRow + 1; i < endRow; i++ ) {
				UtilsFactory::CALC_UTILS.calcIncrementalLCM ( getValue ( i, pivotColumn ) );
			}
			UtilsFactory::CALC_UTILS.convertResultToNegativeValue();
			multiplyUpperTriangle( beginRow, endRow, pivotColumn, UtilsFactory::CALC_UTILS.getResult() );	
		}

		void eliminateColumnTriangle ( const unsigned long beginRow, const unsigned long endRow ) {

			for ( unsigned long index = beginRow; index < endRow - 1; index++ ) {
				if ( getValue( index, endRow - 1 ) != 0 ) {
					getRow( index ).add( endRow - 1, getRow( endRow - 1 ) );
					getRow( index ).add( getRows() - 1, getColumns(), getRow( endRow - 1 ) );
				}
			}
		}

		void eliminateColumnRectangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long beginColumn ) {

			for ( unsigned long index = beginRow + 1; index < endRow; index++ ) {
				if ( getValue( index, beginColumn ) != 0 ) {
					getRow( index ).add( beginColumn, getColumns(), getRow( beginRow ) );
				}
			}
		}

		void eliminateColumnDown ( const unsigned long beginRow, const unsigned long endRow, const unsigned long column ) {

#ifdef __KEEP_MATRIX_NUMBERS_LOW__
            bool simplify_previous = false;
#endif

#ifndef __MINIMIZE_MUL_OPERATIONS__
            const CalculationVector & first_row = getRow( beginRow );
#else
            CalculationVector & first_row = getRow( beginRow );

            UtilsFactory::CALC_UTILS.startIncrementalLCM( first_row[ column ] );

            for ( unsigned long r_index = beginRow + 1; r_index < endRow; r_index++ ) {
                UtilsFactory::CALC_UTILS.calcIncrementalLCM( getValue( r_index, column ) );
            }

            UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() );
            UtilsFactory::CALC_UTILS.divideRegisterS( first_row[ column ] );

            first_row.multiply( column, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );

    #ifdef __KEEP_MATRIX_NUMBERS_LOW__
            UtilsFactory::CALC_UTILS.loadRegisterT( UtilsFactory::CALC_UTILS.getResult() );
    #endif
#endif

			for ( unsigned long r_index = beginRow + 1; r_index < endRow; r_index++ ) {
                CalculationVector & current_row = getRow( r_index );
#ifdef __KEEP_MATRIX_NUMBERS_LOW__
                eliminateColumnDown( first_row, current_row, column, getRow( r_index - 1 ), simplify_previous );
#else
                eliminateColumnDown( first_row, current_row, column );
#endif //  __KEEP_MATRIX_NUMBERS_LOW__
			}

#ifdef __KEEP_MATRIX_NUMBERS_LOW__
            if ( simplify_previous && likely( UtilsFactory::CALC_UTILS.getRegisterG() != 0 ) ) {
                getRow( endRow - 1 ).divide( column + 1, getColumns(), UtilsFactory::CALC_UTILS.getRegisterG() );
            }
#endif // __KEEP_MATRIX_NUMBERS_LOW__
		}


#ifdef __KEEP_MATRIX_NUMBERS_LOW__
		void eliminateColumnDown ( const CalculationVector & reference, CalculationVector & current, const unsigned long column, CalculationVector & previous, bool & simplify_previous ) {

		    if ( current.isNonZeroAt( column ) ) {
                mpz_class & current_column_value = current[ column ];

#ifdef __MINIMIZE_MUL_OPERATIONS__
                UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.get_register_t() / current_column_value );
#else
                const mpz_class & reference_column_value = reference[ column ];
                    
                UtilsFactory::CALC_UTILS.calcGCD( UtilsFactory::CALC_UTILS.get_register_s(), reference_column_value, current_column_value );

                UtilsFactory::CALC_UTILS.loadRegisterT( current_column_value );
                UtilsFactory::CALC_UTILS.divideRegisterT( UtilsFactory::CALC_UTILS.get_register_s() ); 

                UtilsFactory::CALC_UTILS.loadRegisterS( reference_column_value / UtilsFactory::CALC_UTILS.get_register_s() );
#endif

                if ( likely( column + 1 < getColumns() ) ) {
                    
                    if ( simplify_previous && likely( UtilsFactory::CALC_UTILS.getRegisterG() != 0 ) ) {
                        mpz_class & current_value = current[ column + 1 ];

                        eliminateColumn( reference[ column + 1 ], current_value, previous[ column + 1 ] );
                        UtilsFactory::CALC_UTILS.startIncrementalGCD( current_value );

                        for ( unsigned int c_index = column + 2; c_index < getColumns(); c_index++ ) {
                            mpz_class & current_value = current[ c_index ];

                            eliminateColumn( reference[ c_index ], current_value, previous[ c_index ] );
                            UtilsFactory::CALC_UTILS.calcIncrementalGCD( current_value );
                        }
                    } else {
                        mpz_class & current_value = current[ column + 1 ];

                        eliminateColumn( reference[ column + 1 ], current_value );
                        UtilsFactory::CALC_UTILS.startIncrementalGCD( current_value );

                        for ( unsigned int c_index = column + 2; c_index < getColumns(); c_index++ ) {
                            mpz_class & current_value = current[ c_index ];

                            eliminateColumn( reference[ c_index ], current_value );
                            UtilsFactory::CALC_UTILS.calcIncrementalGCD( current_value );
                        }
                    }
                }
				current_column_value = 0;

                UtilsFactory::CALC_UTILS.loadRegisterG( UtilsFactory::CALC_UTILS.getResult() );
                simplify_previous = true;
			} else {
                simplify_previous = false;
            }
		}      
#else
		void eliminateColumnDown ( const CalculationVector & reference, CalculationVector & current, const unsigned long column ) {

		    if ( current.isNonZeroAt( column ) ) {
                mpz_class & current_column_value = current[ column ];
                    
#ifdef __MINIMIZE_MUL_OPERATIONS__
                UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() );
                UtilsFactory::CALC_UTILS.divideRegisterS( current_column_value );
#else
                const mpz_class & reference_column_value = reference[ column ];

                mpz_gcd( UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t(), reference_column_value.get_mpz_t(), current_column_value.get_mpz_t() );
                mpz_tdiv_q( UtilsFactory::CALC_UTILS.get_register_t().get_mpz_t(), current_column_value.get_mpz_t(), UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t() );
                mpz_tdiv_q( UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t(), reference_column_value.get_mpz_t(), UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t() );
#endif
                current_column_value = 0;

                std::vector<mpz_class>::const_iterator ref_it = reference.iterator( column + 1 );
                std::vector<mpz_class>::iterator cur_it = current.iterator( column + 1 );
                const std::vector<mpz_class>::const_iterator ref_it_end = reference.iterator( getColumns() );

                while ( likely( ref_it != ref_it_end ) ) {
                    eliminateColumn( *ref_it++, *cur_it++ );
                } 
             }
		} 
#endif // __KEEP_MATRIX_NUMBERS_LOW__

        void eliminateColumn ( const mpz_class & reference, mpz_class & current ) {
            current *= UtilsFactory::CALC_UTILS.get_register_s();
#ifdef __MINIMIZE_MUL_OPERATIONS__
            current -= reference;
#else
            if ( mpz_cmp_ui( reference.get_mpz_t(), 0 ) != 0 ) {
                mpz_submul( current.get_mpz_t(), UtilsFactory::CALC_UTILS.get_register_t().get_mpz_t(), reference.get_mpz_t() );
            }
#endif
        }  

        void eliminateColumn ( const mpz_class & reference, mpz_class & current, mpz_class & previous ) {
            eliminateColumn( reference, current );
            previous /= UtilsFactory::CALC_UTILS.getRegisterG();
        }  

		void eliminateColumnUp ( const unsigned long beginRow, const unsigned long endRow, const unsigned long column ) {
            const unsigned short last_row = endRow - 1;
            CalculationVector & reference = getRow( last_row );
            
#ifdef __KEEP_MATRIX_NUMBERS_LOW_DIAGONAL__
            bool simplify_previous = false;
#endif

#ifdef __MINIMIZE_MUL_OPERATIONS__
            UtilsFactory::CALC_UTILS.startIncrementalLCM( reference[ column ] );

            for ( unsigned long r_index = beginRow; r_index < last_row; r_index++ ) {
                UtilsFactory::CALC_UTILS.calcIncrementalLCM( getValue( r_index, column ) );
            }

            UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() );
            UtilsFactory::CALC_UTILS.divideRegisterS( reference[ column ] );

            reference.multiply( column, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );
#endif

			for ( unsigned long r_index = beginRow; likely( r_index < last_row ); r_index++ ) {
                CalculationVector & current = getRow( r_index );

#ifdef __KEEP_MATRIX_NUMBERS_LOW_DIAGONAL__
                eliminateColumnUp( reference, current, column, r_index, getRow( r_index - 1 ), simplify_previous );
#else
                eliminateColumnUp( reference, current, column, r_index );
#endif 
            }

#ifdef __KEEP_MATRIX_NUMBERS_LOW_DIAGONAL__
            if ( simplify_previous ) {
                CalculationVector & last = getRow( last_row - 1 );
                last.divide( column + 1, getColumns(), UtilsFactory::CALC_UTILS.getRegisterG() );
                last[ column - 1 ] /= UtilsFactory::CALC_UTILS.getRegisterG();
            } 
#endif
		}

#ifdef __KEEP_MATRIX_NUMBERS_LOW_DIAGONAL__
        void eliminateColumnUp ( const CalculationVector & reference, CalculationVector & current, const unsigned long column, const unsigned long r_index , CalculationVector & previous, bool & simplify_previous ) {

		    if ( current.isNonZeroAt( column ) ) {
                mpz_class & current_column_value = current[ column ];

    #ifdef __MINIMIZE_MUL_OPERATIONS__
                UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() / current_column_value );
    #else                   
                const mpz_class & reference_column_value = reference[ column ];

                mpz_gcd( UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t(), reference_column_value.get_mpz_t(), current_column_value.get_mpz_t() );
                mpz_tdiv_q( UtilsFactory::CALC_UTILS.get_register_t().get_mpz_t(), current_column_value.get_mpz_t(), UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t() );
                mpz_tdiv_q( UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t(), reference_column_value.get_mpz_t(), UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t() );
    #endif

                if ( simplify_previous ) {
                    mpz_class & current_value = current[ r_index ];

                    previous[ r_index - 1 ] /= UtilsFactory::CALC_UTILS.getRegisterG();
                    previous[ r_index ] /= UtilsFactory::CALC_UTILS.getRegisterG();

                    current_value *= UtilsFactory::CALC_UTILS.get_register_s();

                    UtilsFactory::CALC_UTILS.startIncrementalGCD( current_value );                

                    for ( unsigned int c_index = r_index + 1; c_index < column; c_index++ ) {
                        mpz_class & current_value = current[ c_index ];

                        current_value *= UtilsFactory::CALC_UTILS.get_register_s();

                        UtilsFactory::CALC_UTILS.calcIncrementalGCD( current_value );
                        previous[ c_index ] /= UtilsFactory::CALC_UTILS.getRegisterG();
                    } 

                    for ( unsigned int c_index = column + 1; c_index < getColumns(); c_index++ ) {
                        mpz_class & current_value = current[ c_index ];

                        eliminateColumn( reference[ c_index ], current_value, previous[ c_index ] );
                        UtilsFactory::CALC_UTILS.calcIncrementalGCD( current_value );
                    }
                } else { 
                    mpz_class & current_value = current[ r_index ];
    
                    current_value *= UtilsFactory::CALC_UTILS.get_register_s();
                    UtilsFactory::CALC_UTILS.startIncrementalGCD( current_value );                

                    for ( unsigned int c_index = r_index + 1; c_index < column; c_index++ ) {
                        mpz_class & current_value = current[ c_index ];

                        current_value *= UtilsFactory::CALC_UTILS.get_register_s();
                        UtilsFactory::CALC_UTILS.calcIncrementalGCD( current_value );
                    } 

                    for ( unsigned int c_index = column + 1; c_index < getColumns(); c_index++ ) {
                        mpz_class & current_value = current[ c_index ];

                        eliminateColumn( reference[ c_index ], current_value );
                        UtilsFactory::CALC_UTILS.calcIncrementalGCD( current_value );
                    }
                }

                current_column_value = 0;
                UtilsFactory::CALC_UTILS.loadRegisterG( UtilsFactory::CALC_UTILS.getResult() );
                simplify_previous = true;
			 } else {
                simplify_previous = false;
             } 
        }
#else
        void eliminateColumnUp ( const CalculationVector & reference, CalculationVector & current, const unsigned long column, const unsigned long r_index ) {

		    if ( current.isNonZeroAt( column ) ) {
                mpz_class & current_column_value = current[ column ];

    #ifdef __MINIMIZE_MUL_OPERATIONS__
                UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() );
                UtilsFactory::CALC_UTILS.divideRegisterS( current_column_value );
    #else  
                const mpz_class & reference_column_value = reference[ column ];

                mpz_gcd( UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t(), reference_column_value.get_mpz_t(), current_column_value.get_mpz_t() );
                mpz_tdiv_q( UtilsFactory::CALC_UTILS.get_register_t().get_mpz_t(), current_column_value.get_mpz_t(), UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t() );
                mpz_tdiv_q( UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t(), reference_column_value.get_mpz_t(), UtilsFactory::CALC_UTILS.get_register_s().get_mpz_t() );
    #endif
                current_column_value = 0;          

                std::vector<mpz_class>::iterator cur_it = current.iterator( r_index );
                const std::vector<mpz_class>::iterator cur_it_end = current.iterator( column );

                while( likely( cur_it != cur_it_end ) ) {
                    mpz_class & current_val = *cur_it++;
                    current_val *= UtilsFactory::CALC_UTILS.get_register_s();
                }

                std::vector<mpz_class>::const_iterator ref_it = reference.iterator( column + 1 );
                const std::vector<mpz_class>::const_iterator ref_it_end = reference.iterator( getColumns() );

                while ( likely( ref_it != ref_it_end ) ) {
                    eliminateColumn( *ref_it++, *++cur_it );
                } 
             } 
        }
#endif

        void eliminateColumn( const CalculationVector & reference, CalculationVector & current, const unsigned int c_start, const unsigned c_end  ) {
            std::vector<mpz_class>::const_iterator ref_it = reference.iterator( c_start );
            std::vector<mpz_class>::iterator cur_it = current.iterator( c_start );
            const std::vector<mpz_class>::const_iterator ref_it_end = reference.iterator( c_end );

            while ( likely( ref_it != ref_it_end ) ) {
                eliminateColumn( *ref_it++, *cur_it++ );
            }
        }

		void multiplyRectangle ( const unsigned long beginRow, const unsigned long pivotColumn, const mpz_class& lcm ) {
			multiplyRectangle( beginRow, getRows(), pivotColumn, getColumns(), lcm );
		}

		void multiplyRectangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn, const mpz_class& lcm ) {
			multiplyRectangle( beginRow, endRow, pivotColumn, getColumns(), lcm );
		}

		void multiplyRectangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn, const unsigned long endColumn, const mpz_class& lcm ) {

			UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
			UtilsFactory::CALC_UTILS.divideRegisterS( getValue( beginRow, pivotColumn ) );
			UtilsFactory::CALC_UTILS.negateRegisterS();
			getRow( beginRow ).multiply( pivotColumn, endColumn, UtilsFactory::CALC_UTILS.getRegisterS() );

			for ( unsigned long row = beginRow + 1; row < endRow; row++ ) {
				if ( getRow( row ).isNonZeroAt( pivotColumn ) && lcm != getValue( row, pivotColumn ) ) {
					UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
					UtilsFactory::CALC_UTILS.divideRegisterS( getValue( row, pivotColumn ) );		
					getRow( row ).multiply( pivotColumn, endColumn, UtilsFactory::CALC_UTILS.getRegisterS() );
				}
			}
		}

		void multiplyUpperTriangle ( const unsigned long beginRow, const unsigned long endRow, const unsigned long pivotColumn, const mpz_class& lcm ) {
			
			for ( unsigned long row = beginRow; row < endRow - 1; row++ ) {
				if ( getRow( row ).isNonZeroAt( pivotColumn ) && lcm != getValue( row, pivotColumn ) ) {
					UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
					UtilsFactory::CALC_UTILS.divideRegisterS( getValue( row, pivotColumn ) );
					getRow( row ).multiply( row, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );
				}	
			}	

			UtilsFactory::CALC_UTILS.loadRegisterS( lcm );
			UtilsFactory::CALC_UTILS.divideRegisterS( getValue( endRow - 1, pivotColumn ) );
			UtilsFactory::CALC_UTILS.negateRegisterS();
			getRow( endRow - 1 ).multiply( endRow - 1, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );
		}

		bool isSolvable () const {
			bool result = true;
				
			for ( unsigned long row = getColumns() - 1; row < getRows(); row++ ) {
				if ( isNonZeroAt( row, getColumns() - 1 ) ) {
					result = false;
					break;
				} 
			}

			return result; 
		}

		bool findPivot ( const unsigned int beginRow, const unsigned int pivotColumn ) {
			bool result = false;

			if ( !isNonZeroAt( beginRow, pivotColumn ) ) {
				for ( unsigned int row = beginRow + 1; row < getRows(); row++ ) {
					if ( isNonZeroAt( row, pivotColumn ) ) {
						swapRows( row, beginRow );
						result = true;
						break;
					}
				}		
			} else {
				result = true;
			}

			return result;
		}

		void setRows ( const unsigned long rows ) {
			mRowLimit = rows;
		}

		void setColumns ( const unsigned long columns ) {
			mColumnLimit = columns;
		}

		public:

		unsigned long getRows () const {
			return mRowLimit;
		}

		unsigned int getColumns () const {
			return mColumnLimit;
		}

        bool empty () const {
            return getRows() == 0;
        }

		void simplify () {
            unsigned long first_row;

            for ( first_row = 0; !isNonZeroAt( first_row, first_row ) && first_row < getRows() - 1; first_row++ );

			UtilsFactory::CALC_UTILS.startIncrementalLCM ( getValue( first_row, first_row ) );

			for ( unsigned long row = first_row + 1; row < getRows() - 1; row++ ) {
					UtilsFactory::CALC_UTILS.calcIncrementalLCM ( getValue( row, row ) );
			}

			for ( unsigned long row = first_row; row < getRows() - 1; row++ ) {
					UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getResult() );
					UtilsFactory::CALC_UTILS.divideRegisterS( getValue( row, row ) );

					getRow( row ).multiply( row, UtilsFactory::CALC_UTILS.getRegisterS() );
					getRow( row ).multiply( getRows() - 1, getColumns(), UtilsFactory::CALC_UTILS.getRegisterS() );
			} 
		}

		const std::string toString () const;
	};
}

#endif // __DD_LINEAR_DEPENDENCE_CALCULATOR__
