/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_CALCULATION_VECTOR__
#define __DD_CALCULATION_VECTOR__

#include "DDVector.h"

#include <vector>
#include <string.h>

#include <gmp.h>
#include <gmpxx.h>

namespace double_description {

	class CalculationVector : public Vector {

		public:
		CalculationVector ( const unsigned int dimension ) :
			Vector( dimension ) {}

		CalculationVector ( const CalculationVector & other ) :
			Vector( other ) {}

		virtual ~CalculationVector () {}

        std::vector<mpz_class>::iterator iterator ( const unsigned int pos ) {
            return data.begin() + pos;
        }

        std::vector<mpz_class>::const_iterator iterator ( const unsigned int pos ) const {
            return data.begin() + pos;
        }

		unsigned int getDimension () const {
			return data.size();
		}
		
		bool isNonZeroAt ( const unsigned int index ) const {
			return mpz_cmp_ui( data[index].get_mpz_t(), 0 ) != 0;
		}

		unsigned int getNonZeroCount () const {
			unsigned int count = 0;

			for ( unsigned int index = 0; index < getDimension(); index++ ) {
				if ( isNonZeroAt( index ) ) {
					count++;
				}
			}
			return count;
		}

		const mpz_class& getValue ( const unsigned int index ) const {
			return data[index];
		}

		void setValue (	const unsigned int index, const mpz_class& value ) {
			data[index] = value;
		}

		void setValue (	const unsigned int index, const long value ) {
			data[index] = value;
		}

		void add ( const unsigned int index, const mpz_class& increment ) {
			data[index] += increment;
		}

		void add ( const unsigned int index, const CalculationVector& vector ) {
			add( index, vector[index] );
		}

		void add ( const unsigned int begin, const unsigned int end, const CalculationVector& vector ) {

			for ( unsigned int index = begin; index < end; index++ ) {
				add( index, vector[index] );
			} 
		}

		void divide ( const unsigned int index, const mpz_class& divisor ) {
			data[index] /= divisor;
		}

		void divide ( const unsigned int begin, const unsigned int end, const mpz_class & divisor ) {

            const std::vector<mpz_class>::iterator begin_it = data.begin() + begin;
            const std::vector<mpz_class>::iterator end_it = data.begin() + end;

			for ( std::vector<mpz_class>::iterator it = begin_it; it != end_it; it++ ) {
                mpz_class & value = *it;
                mpz_tdiv_q( value.get_mpz_t(), value.get_mpz_t(), divisor.get_mpz_t() );
			} 
		}

		void multiply ( const unsigned int index, const  mpz_class& scalar ) {
			data[index] *= scalar;
		}

		void multiply ( const unsigned int begin, const unsigned int end, const mpz_class & scalar ) {

            const std::vector<mpz_class>::iterator begin_it = data.begin() + begin;
            const std::vector<mpz_class>::iterator end_it = data.begin() + end;

			for ( std::vector<mpz_class>::iterator it = begin_it; it != end_it; it++ ) {
                mpz_class & value = *it;
                value *= scalar;
			} 
		}

		void subtract ( const unsigned int index, const  mpz_class& subtractor ) {
			data[index] -= subtractor;
		}

		void subtract ( const unsigned int index, const CalculationVector& vector ) {
			subtract( index, vector[index] );
		}

		void negate ( const unsigned int index ) {
			data[index] = -data[index];	
		}

		const mpz_class& operator[] ( const unsigned int index ) const {
			return getValue( index );
		}

		mpz_class & operator[] ( const unsigned int index ) {
			return data[ index ];
		}

		bool operator== ( const CalculationVector& vector ) const {

			if ( vector.getDimension() != getDimension() )
				return false;

			for ( unsigned int index = 0; index < getDimension(); index++ ) {
				if( vector[index] != getValue( index ))
					return false;
			}

			return true;
		} 

        void operator= ( const CalculationVector& vector ) {

            for ( unsigned int i = 0; i < vector.data.size(); i++ ) {
                data[ i ] = vector.data[ i ];
            }
        }

		void calcScalarProduct ( const CalculationVector& vector, mpz_class& result ) const {
			result = 0;

			for ( unsigned int index = 0; index < getDimension(); index++ ) {
				result += vector[index] * getValue( index );
			}		
		}

        void simplify ( const unsigned short begin, const unsigned short end ) {

            UtilsFactory::CALC_UTILS.startIncrementalGCD( getValue( begin ) );

            for ( unsigned short i = begin + 1; i < end; i++ ) {
                 UtilsFactory::CALC_UTILS.calcIncrementalGCD( data[ i ] );
            }
        
            divide( begin, end, UtilsFactory::CALC_UTILS.getResult() );
        }

		const std::string toString () const;
	};
}

#endif // __DD_CALCULATION_VECTOR__
