
#ifndef __DD_COMBINATIONAL_TEST__
#define __DD_COMBINATIONAL_TEST__

#include "DDTreeRayContainer.h"
#include "DDBitSetTreeNode.h"
#include "DDPopulationTreeNode.h"

namespace double_description {

	template <class RAY_TYPE, class TREE_TYPE>
	class CombinationalTest {

		private:
		const TREE_TYPE * mPositiveRoot;
		const TREE_TYPE * mNegativeRoot;
		const TREE_TYPE * mZeroRoot;

		unsigned short mPatternThreshold;
		unsigned short mInverseThreshold;

		BitList register_A;
		BitList register_B;

		const BitList* mPatternSet;
		unsigned short mPatternSetSize;

		public:
		CombinationalTest (  const unsigned short blocks ) :
			register_A( blocks ) {}

		CombinationalTest ( const TREE_TYPE * const positive, const TREE_TYPE * const negative, const TREE_TYPE * const zero, const unsigned short patternThreshold, const unsigned short inverseThreshold, const unsigned short blocks,
							const BitList * patternSet, const unsigned short patternSetSize ) :
			mPositiveRoot( positive ),
			mNegativeRoot( negative ),
			mZeroRoot( zero ),
			mPatternThreshold( patternThreshold ),
			mInverseThreshold( inverseThreshold ),
			register_A( blocks ),
			register_B( positive->getPositiveHash().getBlocks() ),
			mPatternSet( patternSet ),
			mPatternSetSize( patternSetSize ) {}

#ifdef __POPULATION_TREES__
		bool adjacencyTest ( const BitList & pList, const BitList & nList, const BitSetTreeNode<RAY_TYPE> * positive, const BitSetTreeNode<RAY_TYPE> * negative, const unsigned long minCommonBits ) {

			register_A = pList;
			register_A &= nList;

			if ( register_A.getPopulationCount() >= minCommonBits ) {

				register_B.clear();
				for ( unsigned short index = 0; index < mPatternSetSize; index++ ) {

					if ( mPatternSet[index].getCommonPopCnt( register_A ) >= mPatternThreshold )
						register_B.setBit( 2 * index );

					if ( mPatternSet[index].getInvCommonPopCnt( register_A ) >= mInverseThreshold )
						register_B.setBit( 2 * index + 1 );
				}

				return	( pList.getPopulationCount() == minCommonBits + 1 ) || ( pList.getPopulationCount() == minCommonBits + 1 ) ||
						( adjacencyCheck( mZeroRoot ) && adjacencyCheck( mPositiveRoot, positive ) && adjacencyCheck( mNegativeRoot, negative ));

			} else {
				return false;
			}
		}
#else // __BIT_PATTERN_TREES__

		bool adjacencyTest ( const BitList & pList, const BitList & nList, const BitSetTreeNode<RAY_TYPE> * root, const BitSetTreeNode<RAY_TYPE> * src, const unsigned long minCommonBits ) {

			register_A = pList;
			register_A &= nList;

			if ( register_A.getPopulationCount() >= minCommonBits ) {
				return (  pList.getPopulationCount() == minCommonBits + 1 ) || ( pList.getPopulationCount() == minCommonBits + 1 ) || adjacencyCheck( root, src );
			} else {
				return false;
			}
		}

		bool adjacencyTest ( const BitList & pList, const BitList & nList, const BitSetTreeNode<RAY_TYPE> * root, const unsigned long minCommonBits ) {

			register_A = pList;
			register_A &= nList;

			if ( register_A.getPopulationCount() >= minCommonBits ) {
				return (  pList.getPopulationCount() == minCommonBits + 1 ) || ( nList.getPopulationCount() == minCommonBits + 1 ) || adjacencyCheck( root );
			} else {
				return false;
			}
		}

#endif

		bool adjacencyCheck ( const BitSetTreeNode<RAY_TYPE> * current ) const {

			if ( current->getMap().contains( register_A ) ) {
		
				if ( current->isLeaf() ) {
					return !superSetExists( current->getRays() );
				} else {
					return	adjacencyCheck( current->getSuperSet() ) &&
							adjacencyCheck( current->getSubSet() );
				}
			} else {
				return true;
			}
		}

		bool adjacencyCheck ( const PopulationTreeNode<RAY_TYPE> * current ) const {

			if ( current->getPositiveHash().contains( register_B ) && current->getMap().contains( register_A ) ) {
		
				if ( current->isLeaf() ) {
					return	adjacencyCheck( current->getBitPatternTree() );
				} else {
					return	adjacencyCheck( current->getDoubleOverNode() ) &&
							adjacencyCheck( current->getPatternUnderNode() ) &&
							adjacencyCheck( current->getInverseUnderNode() );
				}
			} else {
				return true;
			}
		}

		bool adjacencyCheck ( const BitSetTreeNode<RAY_TYPE> * current, const BitSetTreeNode<RAY_TYPE> * src ) const {

			if ( current->getMap().contains( register_A ) ) {
		
				if ( current->isLeaf() ) {
					return ( current == src ) ? !multipleSuperSetsExist( current->getRays() ) : !superSetExists( current->getRays() );
				} else {
					return	adjacencyCheck( current->getSubSet(), src ) && 
							adjacencyCheck( current->getSuperSet(), src );
				}
			} else {
				return true;
			}
		}

		bool adjacencyCheck ( const PopulationTreeNode<RAY_TYPE> * current, const PopulationTreeNode<RAY_TYPE> * src ) const {

			if ( current->getPositiveHash().contains( register_B ) && current->getMap().contains( register_A ) ) {
		
				if ( current->isLeaf() ) {
					return ( current == src ) ? !multipleSuperSetsExist( current->getRays() ) : !superSetExists( current->getRays() );
				} else {
					return	adjacencyCheck( current->getDoubleOverNode(), src ) &&
							adjacencyCheck( current->getPatternUnderNode(), src ) &&
							adjacencyCheck( current->getInverseUnderNode(), src );
				}
			} else {
				return true;
			}
		}

		bool adjacencyCheck ( const PopulationTreeNode<RAY_TYPE> * current, const BitSetTreeNode<RAY_TYPE> * src ) const {

			if ( current->getPositiveHash().contains( register_B ) && current->getMap().contains( register_A ) ) {
		
				if ( current->isLeaf() ) {
					return	adjacencyCheck( current->getBitPatternTree(), src );
				} else {
					return	adjacencyCheck( current->getDoubleOverNode(), src ) &&
							adjacencyCheck( current->getPatternUnderNode(), src ) &&
							adjacencyCheck( current->getInverseUnderNode(), src );
				}
			} else {
				return true;
			}
		}

#ifdef __POPULATION_TREES__
		bool adjacencyCheck ( const TREE_TYPE * current, const TREE_TYPE * positive, const TREE_TYPE * negative ) const {

			if ( current->getPositiveHash().contains( register_B ) ) {
				if ( current->isLeaf() ) {

					bool needs_check = true;

					const TreeRayContainer<RAY_TYPE>* leaf_container = current->getContainer();
					const BitList & leaf_positive_union = leaf_container->getPositiveUnion();
					const BitList & leaf_negative_union = leaf_container->getNegativeUnion();
					const BitList & leaf_zero_union = leaf_container->getZeroUnion();

					if ( leaf_zero_union.contains( register_A ) ) {
						needs_check = !superSetExists( leaf_container->getZeroRays() );
					} 

					if ( needs_check && leaf_positive_union.contains( register_A ) ) {
						needs_check = needs_check &&
						( current == positive ) ? !multipleSuperSetsExist( leaf_container->getPositiveRays() ) : !superSetExists( leaf_container->getPositiveRays() );
					}

					if ( needs_check && leaf_negative_union.contains( register_A ) ) {
						needs_check = needs_check &&
						( current == negative ) ? !multipleSuperSetsExist( leaf_container->getNegativeRays() ) : !superSetExists( leaf_container->getNegativeRays() );
					} 
			
					return needs_check;
				} else {
					return	adjacencyCheck( current->getDoubleOverNode(), positive, negative ) &&
							adjacencyCheck( current->getPatternUnderNode(), positive, negative ) &&
							adjacencyCheck( current->getInverseUnderNode(), positive, negative );

				}
			} else {
				return true;
			}
		}

#endif // __POPULATION_TREES__

		bool superSetExists ( const std::vector<const RAY_TYPE*> & rays ) const {

			for ( unsigned long index = 0; index < rays.size(); index++ ) {
				if ( rays[index]->getActiveSet().contains( register_A ) ) {
					if ( rays[index]->getActiveSet() != register_A ) {
						return true;
					}
				}
			}

			return false;
		} 

		bool multipleSuperSetsExist ( const std::vector<const RAY_TYPE*> & rays ) const {

			bool found = false;

			for ( unsigned long index = 0; index < rays.size(); index++ ) {
				if ( rays[index]->getActiveSet().contains( register_A ) ) {
					if ( found ) {
						return true;
					} else {
						found = true;
					}
				}
			}

			return false;
		} 
	};
}

#endif // _DD_COMBINATIONAL_TEST__
