/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_DUAL_SET_RAY__
#define __DD_DUAL_SET_RAY__

#include "DDBitList.h"
#include "DDRay.h"

#include "DDLinearDependenceEntry.h"
#include "DDLinearDependenceCache.h"

namespace double_description {

#ifdef __COMPRESSION__
	class InternalDependenceMap {

		private:
		std::vector<LinearDependenceEntry*> mCacheMap;
		unsigned char* mIndexToCoeff;

		public:
		InternalDependenceMap ( const unsigned char size ) :
			mCacheMap( 1, NULL) {
			mIndexToCoeff = new unsigned char[size];

			for ( unsigned char i = 0; i < size; i++ ) {
				mIndexToCoeff[i] = 0;
			}
		}
		~InternalDependenceMap () {
			delete[] mIndexToCoeff;
		}

		const LinearDependenceEntry* getEntry ( const unsigned char index ) const {
			return mCacheMap[mIndexToCoeff[index]];
		}

		void setEntry ( LinearDependenceEntry* entry, const unsigned char index ) {
			mIndexToCoeff[index] = mCacheMap.size();
			mCacheMap.push_back( entry );
		}
	};
#endif // __COMPRESSION__

	class DualSetRay : public Ray {

		private:
		const BitList mPrimaryActiveSet;
		BitList mSecondaryActiveSet;
		
		LinearDependenceEntry** mInternalMap;

		public:

		DualSetRay ( const RepresentationVector & vector, mpz_class* const coeffs, LinearDependenceEntry** const internalDependenceMap, const BitList & primaryActiveSet, const BitList & secondaryActiveSet ) :
			Ray( vector, vector.getDimension() - 1, coeffs ),
			mPrimaryActiveSet( primaryActiveSet ),
			mSecondaryActiveSet( secondaryActiveSet ),
			mInternalMap( internalDependenceMap ) {}

		DualSetRay ( const unsigned int dimension, const BitList & primaryActiveSet,  const BitList & secondaryActiveSet, LinearDependenceEntry** const internalDependenceMap, const unsigned int plane, const unsigned int totalPlanes ) :
			Ray( dimension, plane, totalPlanes - ( plane + 1 ) ),
			mPrimaryActiveSet( primaryActiveSet ),
			mSecondaryActiveSet( secondaryActiveSet ),
			mInternalMap( internalDependenceMap ) {}

		~DualSetRay () {

			for ( unsigned int block = 0, prefix = 0; block < mSecondaryActiveSet.getBlocks(); block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = mSecondaryActiveSet[ block ];

				while ( block_data ) {
					unsigned int suffix = __builtin_ctzl( block_data );
					unsigned long mask = __SINGLE_BIT__ << suffix;
					unsigned int index = prefix + suffix;

                    mInternalMap[ index - getDimension() ]->dereference();
					block_data ^= mask;
				}
			}

			delete[] mInternalMap;
		}

		const BitList & getPrimaryActiveSet () const {
			return mPrimaryActiveSet;
		}

		const BitList & getSecondaryActiveSet () const {
			return mSecondaryActiveSet;
		}

        const BitList & getMinimalActiveSet () const {
            return mPrimaryActiveSet;
        }

		const BitList getActiveSet () const {
			return mPrimaryActiveSet | mSecondaryActiveSet;
		}

		void copyActiveSetData ( unsigned long * buffer, const unsigned short blocks ) const {

			buffer[0] = mPrimaryActiveSet[0] | mSecondaryActiveSet[0];

			for ( unsigned short block = 1;  block < blocks; block++ ) {
				buffer[block] = mPrimaryActiveSet[block] | mSecondaryActiveSet[block];
			}
		}

		void copyActiveSetData ( unsigned long * buffer ) const {

			buffer[0] = mPrimaryActiveSet[0] | mSecondaryActiveSet[0];

#ifdef __DOUBLE_BLOCK_OPT__
			buffer[1] = mPrimaryActiveSet[1] | mSecondaryActiveSet[1];
#endif //__DOUBLE_BLOCK_OPT__

#ifdef __DOUBLE_BLOCK_OPT__
			for ( unsigned short block = 2; __builtin_expect( block < mPrimaryActiveSet.getBlocks(), 0 ); block++ ) {
#else

	#ifdef __SINGLE_BLOCK_OPT__
			for ( unsigned short block = 1; __builtin_expect( ( block < mPrimaryActiveSet.getBlocks() ), 0 ); block++ ) {
	#else
			for ( unsigned short block = 1;  block < mPrimaryActiveSet.getBlocks(); block++ ) {
	#endif // __SINGLE_BLOCK_OPT__

#endif //__DOUBLE_BLOCK_OPT__

				buffer[block] = mPrimaryActiveSet[block] | mSecondaryActiveSet[block];
			}
		}

		void copyActiveSetData ( unsigned long * buffer, const unsigned long * const mask ) const {

			buffer[0] = ( mPrimaryActiveSet[0] | mSecondaryActiveSet[0] ) & mask[0];

#ifdef __DOUBLE_BLOCK_OPT__
			buffer[1] = ( mPrimaryActiveSet[1] | mSecondaryActiveSet[1] ) & mask[1];
#endif //__DOUBLE_BLOCK_OPT__

#ifdef __DOUBLE_BLOCK_OPT__
			for ( unsigned short block = 2; __builtin_expect( block < mPrimaryActiveSet.getBlocks(), 0 ); block++ ) {
#else

	#ifdef __SINGLE_BLOCK_OPT__
			for ( unsigned short block = 1; __builtin_expect( ( block < mPrimaryActiveSet.getBlocks() ), 0 ); block++ ) {
	#else
			for ( unsigned short block = 1;  block < mPrimaryActiveSet.getBlocks(); block++ ) {
	#endif // __SINGLE_BLOCK_OPT__

#endif //__DOUBLE_BLOCK_OPT__

				buffer[block] = ( mPrimaryActiveSet[block] | mSecondaryActiveSet[block] ) & mask[block];
			}
		}

		const LinearDependenceEntry* getLinearDependenceEntry ( const unsigned int index ) const {
			return mInternalMap[index-getDimension()];
		} 

		LinearDependenceEntry* link ( const unsigned int index ) const {
            LinearDependenceEntry * link = mInternalMap[ index - getDimension() ]; 
            link->reference();
			return link;
		} 

		void setLinearDependenceEntry ( const unsigned int index, LinearDependenceEntry * const entry ) {
			mInternalMap[index-getDimension()] = entry;
			mSecondaryActiveSet.setBit( index );
		} 

        bool degenerate () const {
            return mSecondaryActiveSet.isEmpty();
        }

		const std::string toString () const;
	};
}

#endif // __DD_DUAL_SET_RAY__
