/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "DDDualSetRayGenerator.h"
#include "DDDualSetRayEnumerator.h"

#include "DDLogging.h"

using namespace double_description;

RepresentationVector DualSetRayGenerator::getInitialVector ( const Matrix & base, const unsigned long positivePlane, BitList & primaryActiveSet ) const {
		
    ExtremeRayUncompressor decompressor( base.getRows(), base.getColumns(), BitList::getBlocks( base.getRows() ));
	RepresentationVector vector( base.getColumns() );

	primaryActiveSet.setBits( 0, base.getRows() );
	primaryActiveSet.unsetBit( positivePlane );

    decompressor.calcExtremeRay( base, primaryActiveSet, vector );

	if ( vector * mSystem[positivePlane] < 0 ) {
	    vector.negate();
	}

	vector.simplify();

	return vector;
}

mpz_class* DualSetRayGenerator::getPlaneCoeffs ( const RepresentationVector & vector, const BitList & primaryActiveSet, BitList & secondaryActiveSet, LinearDependenceEntry** & internalMap ) {

	unsigned long index;
	mpz_class* coeffs = new mpz_class[ mSystem.getRows() - mSystem.getColumns() ];

	for ( index = mSystem.getColumns(); index < mSystem.getRows(); index++ ) {
		vector.calcScalarProduct( mSystem[index], coeffs[ index - mSystem.getColumns() ] );
		if ( coeffs[ index - mSystem.getColumns() ] == 0 ) {
			secondaryActiveSet.setBit( index );
			internalMap[index-mSystem.getColumns()] = mDependenceCache.acquireLinearDependenceEntry( primaryActiveSet, index );
		} else if ( coeffs[ index - mSystem.getColumns() ] < 0 ) {
			break;
		}
	}

	for ( ; index < mSystem.getRows(); index++ ) {
		vector.calcScalarProduct( mSystem[index], coeffs[ index - mSystem.getColumns() ] );
	}

	mDependenceCache.update();

	return coeffs;
}

void DualSetRayGenerator::calculateBase () {

	Matrix base = mSystem.getSubMatrix( 0, mSystem.getColumns() );
/*
	for ( unsigned long row = 1; row < base.getRows(); row++ ) {
		for ( unsigned long column = 0; column < base.getColumns() - row; column++ ) {

			if ( base.isNonZeroAt( base.getRows() - row - 1, column ) ) {
				base.setValue( base.getRows() - row - 1, column, base[base.getRows() - row - 1][column] * base[base.getRows() - row][base.getRows() - row] );
			}
		}
	} 
*/
	for ( unsigned long row = 0; row < base.getRows(); row++ ) {

		BitList primary_active_set( BitList::getBlocks( mSystem.getRows() ) );
		BitList secondary_active_set( BitList::getBlocks( mSystem.getRows() ) );	

		LinearDependenceEntry** internal_map = new LinearDependenceEntry*[ mSystem.getRows() - mSystem.getColumns() ];

		RepresentationVector vector = getInitialVector( base, row, primary_active_set );
		mpz_class* coeffs = getPlaneCoeffs( vector, primary_active_set, secondary_active_set, internal_map );
		DualSetRay* ray = new DualSetRay( vector, coeffs, internal_map, primary_active_set, secondary_active_set );
	
		mCache.insert( ray, 0 );
	}
}

const struct StepStats DualSetRayGenerator::enumerate ( const unsigned long plane, BitPatternTreeCacheEntry<DualSetRay> & cache_entry ) {
    DualSetRayEnumerator enumerator( mSystem, plane, cache_entry, mCache, mDependenceCache, mProver );
    const struct StepStats stats = enumerator.enumerate();

    mDependenceCache.releaseUnusedReferences( plane );

    return stats;
}
