/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_EXTREME_RAY_ENUMERATOR__
#define __DD_EXTREME_RAY_ENUMERATOR__

#include <vector>
#include <sstream>
#include <algorithm>
#include <math.h>

#include "DDStats.h"

#include "DDOptions.h"

#include "DDBitList.h"
#include "DDBitCountTable.h"
#include "DDSingleSetRay.h"
#include "DDBitSelector.h"

#include "DDMostCommonBitFirstStrategy.h"
#include "DDMostAverageBitFirstStrategy.h"
#include "DDLeastAverageBitFirstStrategy.h"
#include "DDLeastCommonBitFirstStrategy.h"
#include "DDFirstAvailableBitStrategy.h"
#include "DDNarrowQueryBitsNeutralizationStrategy.h"
#include "DDVerQueryBitsNeutralizationStrategy.h"

#include "DDIntermediateBitSetNode.h"
#include "DDLeafNode.h"
#include "DDVerificationCondition.h"

#include "DDExtVerificationCondition.h"

#include "DDRayReferenceContainer.h"

#include "DDBitSplitVisitor.h"
#include "DDSingleBitSplitVisitor.h"

#include "DDBinaryStaticVantagePointVisitor.h"
#include "DDBinaryDynamicVantagePointVisitor.h"

#include "DDBitFrequencyModifier.h"
#include "DDZeroFrequencyModifier.h"

#include "DDDegeneracyCalcVisitor.h"
#include "DDRayRefReleaseVisitor.h"

#include "DDPreliminaryCondition.h"
#include "DDVantagePointPreliminaryCondition.h"

#include "DDRayCache.h"

namespace double_description {
	
	template <class RAY_TYPE>
	class ExtremeRayEnumerator {

        protected:
        const Matrix & mSystem;

        const unsigned short mMinCommonBits;
        const unsigned short mPlane;
        const unsigned short mBlocks;

        const RayReferenceContainer & mPositiveNonDegEntries;
        const RayReferenceContainer & mPositiveDegEntries;
        const RayReferenceContainer & mNegativeEntries;
        const RayReferenceContainer & mVerificationEntries;

        RayCache<RAY_TYPE, BitPatternTreeCacheEntry<RAY_TYPE> > & mCache;

        BitSplitVisitor * createBitSplitVisitor ( const TreeOptions & options,
                                                  const RayReferenceContainer & container,
                                                  const RayReferenceContainer & query_data ) const {

            BitSplitVisitor * visitor;

            if ( options.strategy() == BINARY_STATIC_VANTAGE_POINT ) {
                if ( container.pop_cnt_range() != 0 ) {
                    std::vector<BitList> points( 3, BitList( mBlocks, ULONG_MAX ) );

                    visitor = new BinaryStaticVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), points );
                    visitor->setSuccessor( new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() ) );
                } else {
                    visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
                }
            } else if ( options.strategy() == BINARY_DYNAMIC_VANTAGE_POINT ) {
                visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
            } else {
                BitSelectStrategy * select_strategy;

                switch ( options.strategy() ) {
                    case BPT_MOST_COMMON_BIT_FIRST:
                        select_strategy = new MostCommonBitFirstStrategy( query_data, mPlane );
                        break;
                    case BPT_LEAST_COMMON_BIT_FIRST:
                        select_strategy = new LeastCommonBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_MOST_AVERAGE_BIT_FIRST:
                        select_strategy = new MostAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_LEAST_AVERAGE_BIT_FIRST:
                        select_strategy = new LeastAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_QUERY_BITS_NEUTRALIZATION:
                        select_strategy = new NarrowQueryBitsNeutralizationStrategy( container, query_data, mPlane );
                        break;
                    default:
                        select_strategy = new FirstAvailableBitStrategy( mPlane );
                }

                visitor = new SingleBitSplitVisitor( options.max_tree_depth(),
                                                     options.leaf_capacity(),
                                                     *select_strategy );

                delete select_strategy;
            }

            return visitor;
        }

        BitSplitVisitor * createBitSplitVisitor ( const TreeOptions & options,
                                                  const RayReferenceContainer & container,
                                                  const RayReferenceContainer & pos_container,
                                                  const RayReferenceContainer & neg_container ) const {

            BitSplitVisitor * visitor;

            if ( options.strategy() == BINARY_STATIC_VANTAGE_POINT ) {
                if ( container.pop_cnt_range() != 0 ) {
                    std::vector<BitList> points( 3, BitList( mBlocks, ULONG_MAX ) );

                    visitor = new BinaryStaticVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), points );
                    visitor->setSuccessor( new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() ) );
                } else {
                    visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
                }
            } else if ( options.strategy() == BINARY_DYNAMIC_VANTAGE_POINT ) {
                visitor = new BinaryDynamicVantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), options.unique_vp() );
            } else {
                BitSelectStrategy * select_strategy;

                switch ( options.strategy() ) {
                    case BPT_MOST_COMMON_BIT_FIRST:
                        select_strategy = new MostCommonBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_LEAST_COMMON_BIT_FIRST:
                        select_strategy = new LeastCommonBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_MOST_AVERAGE_BIT_FIRST:
                        select_strategy = new MostAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_LEAST_AVERAGE_BIT_FIRST:
                        select_strategy = new LeastAverageBitFirstStrategy( container, mPlane );
                        break;
                    case BPT_QUERY_BITS_NEUTRALIZATION:
                        select_strategy = new VerQueryBitsNeutralizationStrategy( container, pos_container, neg_container, mPlane );
                        break;
                    default:
                        select_strategy = new FirstAvailableBitStrategy( mPlane );
                }

                visitor = new SingleBitSplitVisitor( options.max_tree_depth(),
                                                     options.leaf_capacity(),
                                                     *select_strategy );

                delete select_strategy;
            }

            return visitor;
        }

        void sortData ( RayReferenceContainer & container,
                        const SortType & type ) {
            
            switch ( type ) {
                case MAX_SET_BITS:
                    container.sort<RefComparatorMostBitsFirst>( RefComparatorMostBitsFirst() );
                    break;
                case MIN_SET_BITS:
                    container.sort<RefComparatorLeastBitsFirst>( RefComparatorLeastBitsFirst() );
                    break;
                case MAX_SET_BITS_LEX:
                    container.sort<RefComparatorMostBitsFirstLex>( RefComparatorMostBitsFirstLex() );
                    break;
                case MIN_SET_BITS_LEX:
                    container.sort<RefComparatorLeastBitsFirstLex>( RefComparatorLeastBitsFirstLex() );
                    break;
                case MAX_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMaxFrequency>( RefComparatorMaxFrequency() );
                    break;
                case MIN_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMinFrequency>( RefComparatorMinFrequency() );
                    break;
                case MAX_SET_BITS_MAX_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMaxSetBitsMaxFrequency>( RefComparatorMaxSetBitsMaxFrequency() );
                    break;   
                case MAX_SET_BITS_MIN_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMaxSetBitsMinFrequency>( RefComparatorMaxSetBitsMinFrequency() );
                    break;       
                case MIN_SET_BITS_MIN_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMinSetBitsMinFrequency>( RefComparatorMinSetBitsMinFrequency() );
                    break; 
                case MIN_SET_BITS_MAX_BIT_FREQUENCY:
                    calcFrequency( container, 1 );
                    container.sort<RefComparatorMinSetBitsMaxFrequency>( RefComparatorMinSetBitsMaxFrequency() );
                    break; 
                case NONE:
                    /* Do nothing here */
                    break;    
            }
        } 

        void calcFrequency ( RayReferenceContainer & container,
                             const bool value ) {

            BitCountVisitor bit_counter( mPlane );
            container.iterate( bit_counter );
            if ( value ) {
                const BitFrequencyModifier modifier( bit_counter.population());
                container.modify( modifier );
            } else {
                const ZeroFrequencyModifier modifier( bit_counter.population(), container.size() );
                container.modify( modifier );
            }
        }

        void calcFrequency ( RayReferenceContainer & container,
                             RayReferenceContainer & query_data,
                             const bool value ) const {

            BitCountVisitor bit_counter( mPlane );
            query_data.iterate( bit_counter );
            if ( value ) {
                const BitFrequencyModifier modifier( bit_counter.population());
                container.modify( modifier );
            } else {
                const ZeroFrequencyModifier modifier( bit_counter.population(), container.size() );
                container.modify( modifier );
            }
        }

        const TreeNode * createNarrowingTree ( const RayReferenceContainer & first,
                                               const RayReferenceContainer & second,
                                               const TreeOptions & options ) {
         
            TreeNode * tree;

            if ( first.size() > options.leaf_capacity() ) {

            //    sortData ( first, options.data_sort() );

                BitSplitVisitor * split_visitor = createBitSplitVisitor( options, first, second );
    #ifdef __USE_TEMPLATES__
                if ( split_visitor->supportsExtendedCondition() ) {
                   IntermediateBitSetNode<VantagePointPreliminaryCondition> * root = new IntermediateBitSetNode<VantagePointPreliminaryCondition>( first.union_map(),
                                                                                         VantagePointPreliminaryCondition( first.union_map(), mMinCommonBits, BitList( mBlocks )/*, Interval(0, 0), Interval( first.min_pop_cnt(), first.max_pop_cnt() )*/ ),
                                                                                         split_visitor->getSplitFactor() );
                   root->partition( first, *split_visitor );
                   tree = root;
                } else {
                   IntermediateBitSetNode<PreliminaryCondition> * root = new IntermediateBitSetNode<PreliminaryCondition>( first.union_map(),
                                                                             PreliminaryCondition( first.union_map(), mMinCommonBits ),
                                                                             split_visitor->getSplitFactor() );
                   root->partition( first, *split_visitor );   
                   tree = root;                 
                }
    #else // Use polymorphism
			    IntermediateBitSetNode * root = new IntermediateBitSetNode( first.union_map(),
                                                                             new PreliminaryCondition( first.union_map(), mMinCommonBits ),
                                                                             split_visitor->getSplitFactor() );
                root->partition( first, *split_visitor );
                tree = root;
    #endif

                delete split_visitor;
            } else {
    #ifdef __USE_TEMPLATES__
                tree = new LeafNode<PreliminaryCondition>( first,
                                                           PreliminaryCondition( first.union_map(), mMinCommonBits ));
    #else // Use polymorphism
                tree = new LeafNode( first,
                                     new PreliminaryCondition( first.union_map(), mMinCommonBits ));
    #endif
            }

            return tree;
        }

        const TreeNode * createVerTree () const {

                const TreeNode * verificator;
                const Options & options = Options::current();

                if ( !mVerificationEntries.empty() ) {
                    BitSplitVisitor * zero_visitor = createBitSplitVisitor( options.verification(),
                                                                            mVerificationEntries,
                                                                            mPositiveDegEntries,
                                                                            mNegativeEntries );
    #ifdef __USE_TEMPLATES__
                    if ( zero_visitor->supportsExtendedCondition() ) {
                        const VantagePointPreliminaryCondition empty_cond( mVerificationEntries.union_map(), mMinCommonBits, BitList( mBlocks ) );
                        IntermediateBitSetNode<VantagePointPreliminaryCondition> * zero_tree = new IntermediateBitSetNode<VantagePointPreliminaryCondition>( mVerificationEntries.union_map(),
                                                                                                                                                             empty_cond,
                                                                                                                                                             zero_visitor->getSplitFactor() );
                        zero_tree->partition( mVerificationEntries, *zero_visitor );
                        verificator = zero_tree;   
                    } else {
                        IntermediateBitSetNode<PreliminaryCondition> * zero_tree = new IntermediateBitSetNode<PreliminaryCondition>( mVerificationEntries.union_map(),
                                                                                                                                     PreliminaryCondition( mVerificationEntries.union_map(), mMinCommonBits ),
                                                                                                                                     zero_visitor->getSplitFactor() ); 
                        zero_tree->partition( mVerificationEntries, *zero_visitor );  
                        verificator = zero_tree;        
                    }
    #else // Use polymorphism
                    IntermediateBitSetNode * zero_tree = IntermediateBitSetNode( mVerificationEntries.union_map(), new PreliminaryCondition( mVerificationEntries.union_map(), mMinCommonBits ), 2 );
                    zero_tree->partition( mVerificationEntries, *zero_visitor );
                    verificator = zero_tree;      
    #endif
                    
                    delete zero_visitor;

                } else {
    #ifdef __USE_TEMPLATES__
                    verificator = & LeafNode<PreliminaryCondition>::EMPTY_NODE;
    #else
                    verificator = LeafNode::EMPTY_NODE;
    #endif
                }

                return verificator;
        }

        virtual unsigned long enumerate_pos_deg ( const TreeNode * const pos_deg_tree ) = 0; 
        virtual unsigned long enumerate_pos_nondeg ( const TreeNode * const pos_nondeg_tree ) = 0;

        public:
        ExtremeRayEnumerator ( const Matrix & system,
                               const unsigned short plane,
                               const BitPatternTreeCacheEntry<RAY_TYPE> & entry,
                               RayCache<RAY_TYPE, BitPatternTreeCacheEntry<RAY_TYPE> > & cache ) :
            mSystem( system ),
            mMinCommonBits( system.getColumns() - 2 ),
            mPlane( plane ),
            mBlocks( BitList::getBlocks( system.getRows() )),
            mPositiveNonDegEntries( entry.positive_nondeg() ),
            mPositiveDegEntries( entry.positive_deg() ),
            mNegativeEntries( entry.negative() ),
            mVerificationEntries( entry.degenerate() ),
            mCache( cache ) {}

		const struct StepStats enumerate () {

            struct StepStats stats;

            if ( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() > 0 && !mNegativeEntries.empty() ) {

                const Options & options = Options::current();

           //     sortData ( mVerificationEntries, options.verification().data_sort() );

    #ifdef __STATS_OUTPUT__
                PreliminaryCondition::start_new_stat_session();
                VerificationCondition::start_new_stat_session();
    #endif
        
                if ( !mPositiveDegEntries.empty() ) {                    
                    const TreeNode * pos_deg_tree = createNarrowingTree( mPositiveDegEntries, mNegativeEntries, options.deg_narrowing() ); 

                    stats.verified = enumerate_pos_deg( pos_deg_tree );

                    delete pos_deg_tree;
                }

                if ( !mPositiveNonDegEntries.empty() ) {
                    const TreeNode * pos_nondeg_tree = createNarrowingTree( mPositiveNonDegEntries, mNegativeEntries, options.non_deg_narrowing() );

                    enumerate_pos_nondeg( pos_nondeg_tree );

                    delete pos_nondeg_tree;
                }

    #ifdef __STATS_OUTPUT__
                stats.elimOps = PreliminaryCondition::NARROWING_OPS;
                double n_log_n_complexity = mNegativeEntries.size() * log2( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() );
                double quad_complexity = mNegativeEntries.size() * ( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() );
                stats.elimComplexity = ( stats.elimOps - n_log_n_complexity ) / ( quad_complexity - n_log_n_complexity ) + 1; 

                stats.verificationCandidates = VerificationCondition::CANDIDATES;    
                stats.verOps = VerificationCondition::OPS + PreliminaryCondition::VERIFICATION_OPS;
                
            
                if ( !mVerificationEntries.empty() && stats.verificationCandidates > 0 ) {
                    double log_n_complexity = log2( mVerificationEntries.size() ) * stats.verificationCandidates;
                    double log_n_3_complexity = log2( mVerificationEntries.size() ) * log2( mVerificationEntries.size() ) * log2( mVerificationEntries.size()  ) * stats.verificationCandidates;
                    double n_complexity = mVerificationEntries.size() * stats.verificationCandidates;
                    stats.verComplexityLg = ( stats.verOps - log_n_complexity )/ ( log_n_3_complexity - log_n_complexity ) + 1; 
                    stats.verComplexityLin = ( stats.verOps - log_n_complexity )/ ( n_complexity - log_n_complexity ) + 1;        
                }

                DegeneracyCalcVisitor degeneracy_visitor( mMinCommonBits + 1 );

                if ( !mPositiveDegEntries.empty() || !mPositiveNonDegEntries.empty() ) {
                    mPositiveDegEntries.iterate( degeneracy_visitor );
                    stats.degeneracyPos = degeneracy_visitor.degeneracy() * mPositiveDegEntries.size() /( mPositiveDegEntries.size() + mPositiveNonDegEntries.size() );
                } else {
                    stats.degeneracyPos = ULONG_MAX;
                }

                if ( !mNegativeEntries.empty() ) {
                    mNegativeEntries.iterate( degeneracy_visitor );
                    stats.degeneracyNeg = degeneracy_visitor.degeneracy();   
                } else {
                    stats.degeneracyNeg = ULONG_MAX;
                }
    #endif
            }

			return stats;
		}
    };
}

#endif // __DD_EXTREME_RAY_ENUMERATOR__
