/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_EXTREME_RAY_GENERATOR__
#define __DD_EXTREME_RAY_GENERATOR__

#include "DDMatrix.h"
#include "DDRayCache.h"
#include "DDStats.h"
#include "DDLogging.h"
#include "DDOutput.h"

#include "DDExtremeRayUncompressor.h"
#include "DDBitPatternTreeCacheEntry.h"

namespace double_description {

    template <class RAY_TYPE>
	class ExtremeRayGenerator {

		protected:
#ifdef __COMPRESSION__
		const Matrix& mOriginal;
#endif
		const Matrix& mSystem;

		RayCache<RAY_TYPE, BitPatternTreeCacheEntry<RAY_TYPE> > mCache;

        virtual const struct StepStats enumerate ( const unsigned long plane, BitPatternTreeCacheEntry<RAY_TYPE> & cache_entry ) = 0;

		void calculateStep ( const unsigned long plane ) {

	        BitPatternTreeCacheEntry<RAY_TYPE> & cache_entry = mCache.next();
            
	        DD_CONCAT ( "[" << plane - mSystem.getColumns() + 1 << "/" << mSystem.getRows() - mSystem.getColumns() << "] "
                            << "( pos. (deg.)/neg. (deg.)/elim. compl./ver. cand./acc./ver. lg(n)^3/ver. n/extr. rays) = " )

	        START_TIMER( STEP_TIME )

            const struct StepStats stats = enumerate( plane, cache_entry ); 

            DD_CONCAT( cache_entry.getPositiveCount() << "(" )
            if ( stats.degeneracyPos != ULONG_MAX ) {
                DD_CONCAT ( stats.degeneracyPos )
            } else {
                DD_CONCAT ( "-" )
            }

            DD_CONCAT( ")/" << cache_entry.getNegativeCount() << "(" )

            if ( stats.degeneracyNeg != ULONG_MAX ) {
                DD_CONCAT ( stats.degeneracyNeg )
            } else {
                DD_CONCAT ( "-" )
            }

	        DD_CONCAT ( ")/" << stats.elimComplexity << "/" << stats.verificationCandidates << "/" << stats.verified << "/" )

            if ( stats.verComplexityLg != 0.0 ) {
                DD_CONCAT ( stats.verComplexityLg )
            } else {
                DD_CONCAT ( "-" )
            }

            DD_CONCAT ( "/" )

            if ( stats.verComplexityLin != 0.0 ) {
                DD_CONCAT ( stats.verComplexityLin )
            } else {
                DD_CONCAT ( "-" )
            }

            DD_CONCAT( "/" )

        #ifdef __STATS_OUTPUT__
            if ( Options::current().plot_required() && stats.elimComplexity > 0 ) {
                LOG_LINE( COMPLEXITY_PLOT, ( plane - mSystem.getColumns() ) << " " << stats.elimComplexity << " " << stats.verComplexityLg << " " << stats.verComplexityLin )
            }
        #endif

            mCache.pop();

	        STOP_TIMER( STEP_TIME )

	        DD_OUTPUT( mCache.getExtremeRays() << ", dt=" << TIMER_VALUE( STEP_TIME ) << "ms" )
        }

        public:
#ifdef __COMPRESSION__
		ExtremeRayGenerator ( const Matrix& original, const Matrix& system, const Matrix& transformator ) :
             mOriginal( original ),
#else
		ExtremeRayGenerator ( const Matrix& system, const Matrix& transformator ) :
#endif
	    mSystem( system ),
	    mCache( system, transformator ) {}

        virtual void calculateBase () = 0;

        void run () {
	        for ( unsigned long plane = mSystem.getColumns(); plane < mSystem.getRows(); plane++ ) {
		        calculateStep( plane );
	        }
        }

        template <class VECTOR>
        void writeExtremeRays ( const char * file ) {
	
            Output output( file );
#ifdef __COMPRESSION__
	        ExtremeRayUncompressor decompressor( mSystem.getRows(), mSystem.getColumns(), BitList::getBlocks(mSystem.getRows()));
	        const std::vector<const RAY_TYPE*> & rays = mCache.getCompressedResults();
            VECTOR current( mSystem.getColumns() );
#else
            const std::vector<const RAY_TYPE*> & rays = mCache.getResults();
#endif
            output.write( std::string( "Extreme rays:" ) );

	        for ( typename std::vector<const RAY_TYPE*>::const_iterator it = rays.begin(); it != rays.end(); it++ ) {
#ifdef __COMPRESSION__
		        decompressor.calcExtremeRay( mOriginal, *it, current );
#else
                // @ToDo: Support no calculation without compression!
#endif 
                output.write( current );
                current.clear();
	        }
        }
    };
}

#endif // __DD_EXTREME_RAY_GENERATOR__
