/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_FACE_PROVER__
#define __DD_FACE_PROVER__

#include "DDBitCountTable.h"
#include "DDBitList.h"
#include "DDBitRegister.h"
#include "DDDualSetRay.h"
#include "DDUtilsFactory.h"

namespace double_description {

	class FaceProver {

		private:
		const unsigned short mPlanes;
		const unsigned short mBlocks;
		const unsigned short mTargetRank;

		short int mClearance;
		unsigned short int mCurrentRank;

		BitRegister mRetainedPrimaryBits;
		BitRegister mEliminatedPrimaryBits;
		BitRegister mCandidateSecondaryBits;
		BitRegister mProvenSecondaryBits;

		BitRegister* mWorkingSet;
		BitRegister mCoeffLoaded;

		mpz_t** mCoeffCache;
        mpz_t gcd;
        mpz_t ci_multiplicator;
        mpz_t si_multiplicator;

		const DualSetRay * mFirstRay;
		const DualSetRay * mSecondRay; 

		void initRegisterSets () {

			const BitList & first_primary_set = mFirstRay->getPrimaryActiveSet();
			const BitList & first_secondary_set = mFirstRay->getSecondaryActiveSet();
			const BitList & second_primary_set = mSecondRay->getPrimaryActiveSet();
			const BitList & second_secondary_set = mSecondRay->getSecondaryActiveSet();

			mCurrentRank = 0;
			mClearance = 0;

			for ( unsigned short block = 0; block < mBlocks; block++ ) {

                const unsigned long first_primary_set_block = first_primary_set[block];
                const unsigned long second_primary_set_block = second_primary_set[block];
                const unsigned long second_secondary_set_block = second_secondary_set[block];
                const unsigned long first_secondary_set_block = first_secondary_set[block];

				mRetainedPrimaryBits[block] = ( first_primary_set_block & ( second_primary_set_block | second_secondary_set_block ) ) | ( first_secondary_set_block & second_primary_set_block );
				mCandidateSecondaryBits[block] = first_secondary_set_block & ( second_primary_set_block | second_secondary_set_block );
				mEliminatedPrimaryBits[block] = first_primary_set_block ^ ( first_primary_set_block & ( second_primary_set_block | second_secondary_set_block ) );
				mProvenSecondaryBits[block] = first_secondary_set_block & second_secondary_set_block;

				mCoeffLoaded[block] = 0;

				mCurrentRank += BitList::BIT_COUNT_TABLE.lookUp( mRetainedPrimaryBits[block] );
				mClearance += BitList::BIT_COUNT_TABLE.lookUp( mProvenSecondaryBits[block] );
			}

			mClearance = ( mCurrentRank + mClearance ) - mTargetRank;
		}

		void initWorkingSet ( const unsigned long & plane ) {

			const LinearDependenceEntry * entry = mFirstRay->getLinearDependenceEntry( plane );
			const BitList & dependence_entry_map = entry->getMap();

			for ( unsigned short block = 0; block < mBlocks; block++ ) {
				mWorkingSet[plane][block] = dependence_entry_map[block] & mEliminatedPrimaryBits[block];
			}
		}

		void initCacheEntries ( const unsigned long & plane ) {
			const LinearDependenceEntry * const entry = mFirstRay->getLinearDependenceEntry( plane );

			for ( unsigned short block = 0, prefix = 0;  block < mBlocks; block++, prefix += __BLOCK_SIZE__ ) {

				unsigned long block_data = mWorkingSet[plane][block];

				while ( block_data ) {
					unsigned long suffix = __builtin_ctzl( block_data );
					const unsigned int index = prefix + suffix;			

					mpz_set( mCoeffCache[plane][index], entry->getCoeff( index ).get_mpz_t() );

					block_data ^= __SINGLE_BIT__ << suffix;
				}	
			}
		}

		void calcKnownDependencies () {

			unsigned long block_data;

			for ( unsigned short block = 0, prefix = 0; block < mBlocks; block++, prefix += __BLOCK_SIZE__ ) {

				block_data = mCandidateSecondaryBits[ block ];

				while ( block_data && mClearance >= 0 ) {
					const unsigned int suffix = __builtin_ctzl( block_data ); 
					unsigned long mask = __SINGLE_BIT__ << suffix;
					const unsigned int index = prefix + suffix;
			
					if ( ( mProvenSecondaryBits[ block ] & mask ) && ( mRetainedPrimaryBits.contains( mFirstRay->getLinearDependenceEntry( index )->getMap() ) || mRetainedPrimaryBits.contains( mSecondRay->getLinearDependenceEntry( index )->getMap() ) ) ) {
						mCandidateSecondaryBits.opXOR( block, mask );
						mClearance--;				
					} else {
						initWorkingSet( index );
					}
	
					block_data ^= mask;
				}
			}
		}

		void calcUnknownDependencies () {

			for ( unsigned short s_block = 0, s_prefix = 0; s_block < mBlocks && mClearance >= 0 && mCurrentRank < mTargetRank; s_block++, s_prefix += __BLOCK_SIZE__ ) {

				while ( mCandidateSecondaryBits[ s_block ] && mClearance >= 0 && mCurrentRank < mTargetRank ) {
					const unsigned int s_suffix = __builtin_ctzl( mCandidateSecondaryBits[ s_block ] ); 
					unsigned long s_mask = __SINGLE_BIT__ << s_suffix;
					const unsigned int s_index = s_prefix + s_suffix;

					if ( mProvenSecondaryBits[ s_block ] & s_mask ) {
						mRetainedPrimaryBits.opOR( s_block, s_mask );
						mProvenSecondaryBits.opXOR( s_block, s_mask );
						mCurrentRank++;
					}

					if ( mCurrentRank < mTargetRank ) {

						unsigned long p_suffix = 0, p_block;
						mEliminatedPrimaryBits.getFirstCommonBit( mWorkingSet[s_index], p_block, p_suffix );
						const unsigned long p_mask = __SINGLE_BIT__ << p_suffix;
						const unsigned long p_index = ( p_block * __BLOCK_SIZE__ ) + p_suffix;

						swapBits( p_index, p_block, p_mask, s_index, s_block, s_mask );
					}
				}			
			}
		}

		void swapBits ( const unsigned long & primaryIndex, const unsigned long & primaryBlock, const unsigned long & primaryMask,
											const unsigned long & secondaryIndex, const unsigned long & secondaryBlock, const unsigned long & secondaryMask ) {

			mEliminatedPrimaryBits.opXOR( primaryBlock, primaryMask );
			mCandidateSecondaryBits.opXOR( secondaryBlock, secondaryMask );

			if ( !( mCoeffLoaded[ secondaryBlock ] & secondaryMask ) ) {
					initCacheEntries( secondaryIndex );
					mCoeffLoaded.opXOR( secondaryBlock, secondaryMask );
			} 

			mWorkingSet[ secondaryIndex ].opXOR( primaryBlock, primaryMask );

			for ( unsigned short block = 0, prefix = 0; block < mBlocks && mClearance >= 0; block++, prefix += __BLOCK_SIZE__ ) {

				unsigned long block_data = mCandidateSecondaryBits[ block ];

				while ( block_data && mClearance >= 0 ) {
					const unsigned int suffix = __builtin_ctzl( block_data ); 
					unsigned long mask = __SINGLE_BIT__ << suffix;
					const unsigned int index = prefix + suffix;

					if ( mWorkingSet[ index ].getOpAND( primaryBlock, primaryMask ) ) {

						if ( !( mCoeffLoaded[ block ] & mask ) ) {
							initCacheEntries( index );
							mCoeffLoaded.opXOR( block, mask );
						} 

						updateWorkingSet( primaryIndex, primaryBlock, primaryMask,
															secondaryIndex, index, block, mask );
					}

					block_data ^= mask;
				}
			}
		}


		void updateWorkingSet ( const unsigned long & primaryIndex, const unsigned long & primaryBlock, const unsigned long & primaryMask,
															const unsigned long & secondaryIndex, const unsigned long & currentIndex, const unsigned long & currentBlock, const unsigned long & currentMask ) {
			
			bool independent = false;

			mWorkingSet[ currentIndex ].opXOR( primaryBlock, primaryMask );

            mpz_gcd( gcd, mCoeffCache[secondaryIndex][primaryIndex], mCoeffCache[currentIndex][primaryIndex] );
            mpz_tdiv_q( ci_multiplicator, mCoeffCache[secondaryIndex][primaryIndex], gcd );
            mpz_tdiv_q( si_multiplicator, mCoeffCache[currentIndex][primaryIndex], gcd ); 

			for ( unsigned short block = 0, prefix = 0;  block < mBlocks; block++, prefix += __BLOCK_SIZE__ ) {

				unsigned long block_data = mWorkingSet[ currentIndex ][ block ] | mWorkingSet[ secondaryIndex ][ block ];

				while ( block_data ) {
					const unsigned int suffix = __builtin_ctzl( block_data ); 
					unsigned long mask = __SINGLE_BIT__ << suffix;
					const unsigned int index = prefix + suffix;

                    mpz_t & cache_currentIndex_index = mCoeffCache[currentIndex][index];

					if ( mWorkingSet[ currentIndex ][ block ] & mask ) {

                        mpz_mul( cache_currentIndex_index, cache_currentIndex_index, ci_multiplicator );

						if ( mWorkingSet[ secondaryIndex ][ block ] & mask ) {
                            const mpz_t & cache_secondaryIndex_index = mCoeffCache[secondaryIndex][index];

                            mpz_submul( cache_currentIndex_index, cache_secondaryIndex_index, si_multiplicator );

							if ( mpz_cmp_ui( cache_currentIndex_index, 0 ) == 0 ) {
								mWorkingSet[ currentIndex ].opXOR( block, mask );
							} else {
								independent = true;
							}
						} else {
							independent = true;
						}
					} else {
                            mpz_neg( cache_currentIndex_index, mCoeffCache[secondaryIndex][index] );
                            mpz_mul( cache_currentIndex_index, cache_currentIndex_index, si_multiplicator ); 

							mWorkingSet[ currentIndex ].opOR ( block, mask );

							independent = true; 
					} 

					block_data ^= mask;
				}
			} 
		
			if ( !independent ) {

				mCandidateSecondaryBits.opXOR( currentBlock, currentMask );
				mProvenSecondaryBits.opOR( currentBlock, currentMask );
				mClearance--;
			}
		}

		public:
		FaceProver ( const unsigned short planes, const unsigned short rank );
		FaceProver ( const FaceProver & other );
		~FaceProver ();

		bool isFace ( const DualSetRay * const positive, const DualSetRay * const negative ) {

			if ( positive->getSecondaryActiveSet().getPopulationCount() > negative->getSecondaryActiveSet().getPopulationCount() ) {
				mFirstRay = negative;
				mSecondRay = positive;
			}	else {
				mFirstRay = positive;
				mSecondRay = negative;
			}

			initRegisterSets ();

			if ( mCurrentRank < mTargetRank ) {
				calcKnownDependencies();
				calcUnknownDependencies();
			}

			return ( mCurrentRank == mTargetRank );
		}

		const BitList & getRetainedPrimaryBits () const {
			return mRetainedPrimaryBits;
		}

		const BitList & getProvenSecondaryBits () const {
			return mProvenSecondaryBits;
		}
	};
}

#endif // __DD_FACE_PROVER__
