/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_INTERMEDIATE_BIT_SET_NODE__
#define __DD_INTERMEDIATE_BIT_SET_NODE__

#include "DDTreeNode.h"
#include "DDLeafNode.h"
#include "DDPreliminaryCondition.h"
#include "DDVerificationCondition.h"

#include "DDRayReferenceContainer.h"

#include "DDBitSplitVisitor.h"

#ifndef __TREE_CHILD_NODES__
#define __TREE_CHILD_NODES__ 2
#endif

namespace double_description {

    #ifdef __USE_TEMPLATES__
    template<class CONDITION_TYPE>
    #endif
	class IntermediateBitSetNode : virtual public TreeNode {

        private:
        const TreeNode * child ( const unsigned short index ) const {
            if ( likely( index < __TREE_CHILD_NODES__ ) ) {
                return mChildBase[ index ];
            } else {
                return mChild[ index - __TREE_CHILD_NODES__ ];
            }
        }

        void insert ( const unsigned short index, const TreeNode * child ) {
            if ( likely( index < __TREE_CHILD_NODES__ ) ) {
                mChildBase[ index ] = child;
            } else {
                mChild[ index - __TREE_CHILD_NODES__ ] = child;
            }
        }

		protected:
    #ifdef __USE_TEMPLATES__    
        const CONDITION_TYPE mCondition;
    #endif
        const TreeNode * mChildBase[ __TREE_CHILD_NODES__ ];
		const TreeNode ** mChild;        
        const unsigned short mChildrenCount;

		TreeNode * createChildNode ( const RayReferenceContainer & candidates,
    #ifdef __USE_TEMPLATES__ 
                                     const CONDITION_TYPE & cond,
    #else // Use polymorphism
                                     const PreliminaryCondition * cond,
    #endif
                                     BitSplitVisitor & visitor,
                                     const BitSplitVisitorState state ) const {

	        if ( state == TERMINATE ) {
    #ifdef __USE_TEMPLATES__
		        return new LeafNode<CONDITION_TYPE>( candidates, cond );
    #else // Use polymorphism
                return new LeafNode( candidates, cond );
    #endif
	        } else {
    #ifdef __USE_TEMPLATES__ 
                IntermediateBitSetNode<CONDITION_TYPE> * node = new IntermediateBitSetNode<CONDITION_TYPE>( candidates.union_map(), cond, visitor.getSplitFactor() );
    #else // Use polymorphism
                IntermediateBitSetNode * node = new IntermediateBitSetNode( candidates.union_map(), cond, visitor.getSplitFactor() );
    #endif
                if ( unlikely( state == CHANGE_VISITOR ) ) {
                    node->partition( candidates, visitor.successor_to_proceed( candidates ) );
                } else {
                    node->partition( candidates, visitor );
                } 
		        return node;
	        }    
        }

		public:

		IntermediateBitSetNode ( const BitList & map,
    #ifdef __USE_TEMPLATES__ 
                                 const CONDITION_TYPE & cond,
    #else // Use polymorphism
                                 const PreliminaryCondition * const cond,
    #endif
                                 const unsigned short children ) :
    #ifdef __USE_TEMPLATES__ 
	        TreeNode( map ),
            mCondition( cond ),
    #else // Use polymorphism
            TreeNode( map, cond ),
    #endif
	        mChildrenCount( children ) {

            if ( unlikely( mChildrenCount > __TREE_CHILD_NODES__ ) ) {
                mChild = new const TreeNode*[ mChildrenCount - __TREE_CHILD_NODES__ ];
            }
        }

		virtual ~IntermediateBitSetNode () {

            if ( unlikely( mChildrenCount > __TREE_CHILD_NODES__ ) ) {
                for ( unsigned short i = 0; i < mChildrenCount - __TREE_CHILD_NODES__ ; i++ ) {
                    delete mChild[i];
                }
                delete[] mChild;
            }
        }

		void partition ( const RayReferenceContainer & candidates,
                         BitSplitVisitor & visitor )  {

            RayReferenceContainer * set = new RayReferenceContainer[ mChildrenCount ];
    #ifdef __USE_TEMPLATES__ 
            CONDITION_TYPE * condition = new CONDITION_TYPE[ mChildrenCount ];
    #else // Use polymorphism
            const PreliminaryCondition ** condition = new const PreliminaryCondition*[ mChildrenCount ];
    #endif
            BitSplitVisitorState * state = new BitSplitVisitorState[ mChildrenCount ];

            visitor.registerOutputSets( set );
            visitor.before( candidates );

            candidates.iterate( visitor );

            for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
    #ifdef __USE_TEMPLATES__ 
                visitor.refineCondition( i, mCondition.getMinCommonBits(), condition[ i ] );
    #else // Use polymorphism
                condition[ i ] = visitor.createNarrowingCondition( i, mCond->getMinCommonBits() );
    #endif
                state[ i ] = visitor.state( i );
            }

            for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
                insert( i, createChildNode( set[ i ], condition[ i ], visitor, state[ i ] ));
            }

            visitor.after();

            delete[] set;
            delete[] condition;
            delete[] state;
        }

		void preliminaryTraverse ( const TreeNode * partner,
									std::vector<struct EdgeCandidate> & candidates ) const {

	        if ( partner->getCondition()->holdsOnFace( mMap ) ) {
                for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
                    partner->preliminaryTraverse( child( i ), candidates );
		        }
	        }
        }

		void preliminaryTraverse (  const RayReference & current,
									std::deque<struct EdgeCandidate> & candidates ) const { 
            
            if (
    #ifdef __USE_TEMPLATES__         
             mCondition.proceed_narrowing( current.activeSet )
    #else // Use polymorphism
             mCond->proceed_narrowing( current.activeSet )
    #endif
            ) {
                for ( short i = mChildrenCount - 1; i >= 0; i-- ) {
                    child( i )->preliminaryTraverse( current, candidates );
                }
	        } 
        }

		void preliminaryTraverse ( const RayReference * refs,
									const unsigned int refsCount,
									const PreliminaryCondition & condition,
									std::vector<struct EdgeCandidate> & candidates ) const {

	        if ( condition.holdsOnFace( mMap ) ) {
                for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
		            child( i )->preliminaryTraverse( refs, refsCount, condition, candidates );
		        }		
	        }
        }

		bool verificationTraverse ( const ExtVerificationCondition & condition ) const {

    #ifdef __USE_TEMPLATES__
            bool redundant = mCondition.isRedundant( condition );
    #else // Use polymorphism
            bool redundant = mCond->isRedundant( condition );
    #endif

	        if ( 
    #ifdef __USE_TEMPLATES__ 
             mCondition.proceed_verification( condition.getMap() ) 
    #else // Use polymorphism
             mCond->proceed_verification( condition.getMap() ) 
    #endif
            ) {
                for ( short i = mChildrenCount - 1; i >= 0 && !redundant; i-- ) {
                    redundant |= child( i )->verificationTraverse( condition );
                } 
	        } 

            return redundant;
        }

        unsigned long getSubNodesCount () const {
            unsigned long count = mChildrenCount;

            for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
                count += child( i )->getSubNodesCount();
            }
            return count;
        }

        unsigned long getRaysCount () const {
            unsigned long count = 0;

            for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
                count += child( i )->getRaysCount();
            }
            return count;
        }

        unsigned long getLeafNodesCount () const {
            unsigned long count = 0;

            for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
                count += child( i )->getLeafNodesCount();
            }

            return count; 
        }

        void enumerateLeafNodes( std::vector<const TreeNode*> & container ) const {
            for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
                child( i )->enumerateLeafNodes( container );
            }
        }

        bool isEmpty () const {
            return false;
        }

        const std::string toString ( const unsigned short depth ) const {
            std::stringstream result;

            const TreeNode * node = this;

            result << "p" << static_cast<const void*>( node ) << " [label=\"" << "|V|=" << getRaysCount() << ", "
    #ifdef __USE_TEMPLATES__
                   << mCondition.toString()          
    #else
                   << mCond->toString()
    #endif    
                   << "\"];\n";

            if ( depth > 0 ) {
                for ( unsigned short i = 0; i < mChildrenCount; i++ ) {
                    result << child( i )->toString( depth - 1 )
                           << "p" << static_cast<const void*>( node ) << "->" << "p" << static_cast<const void*>( child( i ) ) << ";\n";
                }
            }

            return result.str();
        }
	};
}

#endif // __DD_INTERMEDIATE_BIT_SET_NODE__
