
#include "DDIntermediatePopulationNode.h"
#include "DDPopTreePreliminaryCondition.h"
#include "DDLogging.h"
#include <sstream>

using namespace double_description;

IntermediatePopulationNode::IntermediatePopulationNode ( const BitList & map,
                                                            const PreliminaryCondition * const cond,
                                                            const unsigned int samplesCount ) : 
	TreeNode( map, cond, USHRT_MAX ),
	PopTreeNode( map, cond, BitList( BitList::getBlocks( samplesCount ), ULONG_MAX ), BitList( BitList::getBlocks( samplesCount ), ULONG_MAX ), USHRT_MAX ) {}

IntermediatePopulationNode::IntermediatePopulationNode ( const BitList & map,
                                                         const PreliminaryCondition * const cond,
														 const BitList & regPattern,
													     const BitList & invPattern,
                                                         const unsigned short degeneracy ) :
	TreeNode( map, cond, degeneracy ),
	PopTreeNode( map, cond, regPattern, invPattern, degeneracy ) {}

void IntermediatePopulationNode::partition ( const std::vector<struct RayReference> & refs,
												PopTreeCreatorUtils & utils,
												const unsigned int sampleID ) {

	BitList over_map( mMap.getBlocks() );
	std::vector<struct RayReference> over_candidates;
	BitList inv_pattern( mInvPattern );
	inv_pattern.unsetBit( sampleID );
    unsigned short over_degeneracy = 0;

	BitList under_map( mMap.getBlocks() );
	std::vector<struct RayReference> under_candidates;
	BitList reg_pattern( mRegPattern );
	reg_pattern.unsetBit( sampleID );
    unsigned short under_degeneracy = 0;

	BitList double_over_map( mMap.getBlocks() );
	std::vector<struct RayReference> double_over_candidates;
    unsigned short double_over_degeneracy = 0;
	
	for ( unsigned long index = 0; index < refs.size(); index++ ) {

        register const unsigned short degeneracy = refs[index].activeSet->getCommonPopCnt( utils.getCutMap() );

		if ( utils.samples().isRegularOver( *(refs[index].activeSet), sampleID ) ) {
			if ( utils.samples().isInverseOver( *(refs[index].activeSet), sampleID ) ) {
				double_over_candidates.push_back( refs[index] );
				double_over_map |= *( refs[index].activeSet );
                if ( __builtin_expect( degeneracy > double_over_degeneracy, 0 ) ) {
                    double_over_degeneracy = degeneracy;
                }
			} else {
				over_candidates.push_back( refs[index] );
				over_map |= *( refs[index].activeSet );
                if ( __builtin_expect( degeneracy > over_degeneracy, 0 ) ) {
                    over_degeneracy = degeneracy;
                }
			}
		} else {
			under_candidates.push_back( refs[index] );
			under_map |= *( refs[index].activeSet );
            if ( __builtin_expect( degeneracy > under_degeneracy, 0 ) ) {
                under_degeneracy = degeneracy;
            }
		}
	}

	mOver = createChildNode( over_candidates, over_map, mRegPattern, inv_pattern, over_degeneracy, utils, sampleID+1 );	
	mUnder = createChildNode( under_candidates, under_map, reg_pattern, mInvPattern, under_degeneracy, utils, sampleID+1 );
	mDoubleOver = createChildNode( double_over_candidates, double_over_map, mRegPattern, mInvPattern, double_over_degeneracy, utils, sampleID+1 );
}

PopTreeNode * IntermediatePopulationNode::createChildNode ( const std::vector<struct RayReference> & candidates,
													    	const BitList & map,
														    const BitList & regPattern,
														    const BitList & invPattern,
                                                            const unsigned short degeneracy,
														    PopTreeCreatorUtils & utils,
														    const unsigned int sampleID ) const {
	PopTreeNode * result;
    const PreliminaryCondition * cond = new PopTreePreliminaryCondition( map, mCond->getMinCommonBits(), regPattern, invPattern );

	if ( candidates.size() < utils.getMaxElementsPerLeaf() || sampleID >= utils.samples().getSamplesCount() ) {
		result = new PopTreeLeafNode( candidates, map, cond, regPattern, invPattern, degeneracy );
	} else {
		IntermediatePopulationNode* node = new IntermediatePopulationNode( map, cond, regPattern, invPattern, degeneracy );
		node->partition( candidates, utils, sampleID );
		result = node;
	} 

	return result;
}

void IntermediatePopulationNode::preliminaryTraverse ( const TreeNode * partner,
														std::vector<const void*> & candidates ) const {

    if ( partner->getCondition()->holdsOnFace( mMap, mRegPattern, mInvPattern ) ) {
	    partner->preliminaryTraverse( mOver, candidates );
	    partner->preliminaryTraverse( mUnder, candidates );
	    partner->preliminaryTraverse( mDoubleOver, candidates );	
    }
}

void IntermediatePopulationNode::preliminaryTraverse ( const RayReference * refs,
														const unsigned int refsCount,
														const PreliminaryCondition & condition,
														std::vector<const void*> & candidates ) const {

	if ( condition.holdsOnFace( mMap, mRegPattern, mInvPattern ) ) {
		mOver->preliminaryTraverse( refs, refsCount, condition, candidates );
		mUnder->preliminaryTraverse( refs, refsCount, condition, candidates );
		mDoubleOver->preliminaryTraverse( refs, refsCount, condition, candidates );
	} 
}

bool IntermediatePopulationNode::verificationTraverse ( const VerificationCondition & condition ) const {

	if ( condition.holdsOnFace( mMap , mRegPattern, mInvPattern, mDegeneracy ) ) {
		return mDoubleOver->verificationTraverse( condition ) ||
               mOver->verificationTraverse( condition ) ||
               mUnder->verificationTraverse( condition );
	} else {
		return false;
	}
}

PopTreeNode * IntermediatePopulationNode::getNextNonRedundant () {

    if ( mOver->isEmpty() && mUnder->isEmpty() ) {
        return mDoubleOver->getNextNonRedundant();
    } else if ( mOver->isEmpty() && mDoubleOver->isEmpty() ) {
        return mUnder->getNextNonRedundant();
    } else if ( mDoubleOver->isEmpty() && mUnder->isEmpty() ) {
        return mOver->getNextNonRedundant();
    } else {
        return this;
    }
}

void IntermediatePopulationNode::optimize () {
    
    if ( mOver->isRedundant() ) {
        mOver = mOver->getNextNonRedundant();
    }

    if ( mDoubleOver->isRedundant() ) {
        mDoubleOver = mDoubleOver->getNextNonRedundant();
    }

    if ( mUnder->isRedundant() ) {
        mUnder = mUnder->getNextNonRedundant();
    }

    mUnder->optimize();
    mOver->optimize();
    mDoubleOver->optimize();
}

void IntermediatePopulationNode::enumerateLeafNodes( std::vector<const TreeNode*> & container ) const {
    mDoubleOver->enumerateLeafNodes( container );
    mOver->enumerateLeafNodes( container );
    mUnder->enumerateLeafNodes( container );
}

void IntermediatePopulationNode::enumeratePopTreeNodes( std::vector<const PopTreeNode*> & container ) const {

    mDoubleOver->enumeratePopTreeNodes( container );
    mOver->enumeratePopTreeNodes( container );
    mUnder->enumeratePopTreeNodes( container );
}

unsigned long IntermediatePopulationNode::getSubNodesCount () const {

    return 3 + mDoubleOver->getSubNodesCount() + 
                mUnder->getSubNodesCount() +
                mOver->getSubNodesCount();
}

unsigned long IntermediatePopulationNode::getRaysCount () const {

    return  mDoubleOver->getRaysCount() +
                mUnder->getRaysCount() +
                mOver->getRaysCount();
}

unsigned long IntermediatePopulationNode::getLeafNodesCount () const {

    return  mDoubleOver->getLeafNodesCount() +
                mUnder->getLeafNodesCount() +
                mOver->getLeafNodesCount();
}

const std::string IntermediatePopulationNode::toString () const {
    std::stringstream result;

    const TreeNode *node, *over, *under, *double_over;

    node = this;

    over = mOver;
    under = mUnder;
    double_over = mDoubleOver;

    result << "p" << static_cast<const void*>( node ) << " [label=" << getRaysCount() << "];\n"
            << over->toString() << under->toString() << double_over->toString()
            << "p" << static_cast<const void*>( node ) << "->" << "p" << static_cast<const void*>( over ) << " [color=red];\n"
            << "p" << static_cast<const void*>( node ) << "->" << "p" << static_cast<const void*>( under ) << " [color=green];\n"
            << "p" << static_cast<const void*>( node ) << "->" << "p" << static_cast<const void*>( double_over ) << ";\n";

    return result.str();
}
