/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_LINEAR_DEPENDENCE_VECTOR__
#define __DD_LINEAR_DEPENDENCE_VECTOR__

#include <vector>
#include <deque>

#include <iostream>
#include <string.h>
#include "DDBitList.h"

#include "DDLinearDependenceEntry.h"

#include "DDLogging.h"

namespace double_description {

	class LinearDependenceVector {
		
		private:
		unsigned long globalLimit;
		unsigned long mBlocks;

		std::vector<unsigned long> mBitData;

		const unsigned long * p_map_start;
		const unsigned long * p_map_end;

		std::deque<LinearDependenceEntry*> mEntryData;

		unsigned long getGlobalLimit () const {
			return globalLimit;
		}

		public:
		LinearDependenceVector ( const unsigned long dim, const unsigned long blocks ) :
	        globalLimit( dim ),
	        mBlocks( blocks ),
	        p_map_start( NULL ),
	        p_map_end( NULL ) {}

		unsigned long getSize () const {
			return mEntryData.size();
		}
    
		void insert ( const BitList & map ) {
	
			mBitData.push_back( map[0] );
#ifdef __DOUBLE_BLOCK_OPT__
			mBitData.push_back( map[1] );
#endif //__DOUBLE_BLOCK_OPT__

#ifdef __DOUBLE_BLOCK_OPT__
			for ( unsigned short block = 2; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
#else

	#ifdef __SINGLE_BLOCK_OPT__
			for ( unsigned short block = 1; __builtin_expect( ( block < mBlocks ), 0 ); block++ ) {
	#else
			for ( unsigned short block = 1;  block < mBlocks; block++ ) {
	#endif // __SINGLE_BLOCK_OPT__

#endif //__DOUBLE_BLOCK_OPT__
				mBitData.push_back( map[block] );
			} 

			p_map_start = &( mBitData.front() );
			p_map_end = &( mBitData.back() ) + 1;
		}

		void insert ( LinearDependenceEntry * const entry ) {
			mEntryData.push_back( entry );
		}

		const LinearDependenceEntry * getDependenceEntry ( const BitList& candidate ) const {

			return acquireDependenceEntry( candidate );
		}

		LinearDependenceEntry * acquireDependenceEntry ( const BitList& candidate ) const {

			for ( const unsigned long * p_map = p_map_start; __builtin_expect( p_map < p_map_end, 1 ); p_map += mBlocks ) {

				if ( unlikely( candidate.containsUnexpected( p_map ) ) ) {
					return mEntryData[ ( p_map - p_map_start ) / mBlocks ];
				}				
			}

			return NULL;
		}

		const LinearDependenceEntry * getDependenceEntry ( const unsigned long position ) const {
			return mEntryData[ position ];
		}

        void releaseUnusedReferences () {

            mBitData.clear();

            for ( std::deque<LinearDependenceEntry*>::iterator it = mEntryData.begin(); it != mEntryData.end(); ) {
                if ( (*it)->is_referenced() ) {
                    insert( (*it)->getMap() );
                    it++;
                } else {
                    delete *it;
                    it = mEntryData.erase( it );
                }
            }
        }

		const std::string toString () const;
	};

}

#endif // __DD_LINEAR_DEPENDENCE_VECTOR__
