/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "DDMatrix.h"
#include "DDUtilsFactory.h"

#include "DDLogging.h"

#include <sstream>

using namespace double_description;

Matrix::Matrix () {}

Matrix::Matrix ( const unsigned int rows, const unsigned int columns ) {

	for ( unsigned int row_index = 0; row_index < rows; row_index++ ) {
		const CalculationVector row( columns );
		mData.push_back( row );
		mMapping.push_back( row_index );
	}	
}

Matrix::Matrix ( const unsigned int rows, const unsigned int columns, const std::vector<unsigned int> & mapping ) :
	mMapping( mapping ) {

	for ( unsigned int row_index = 0; row_index < rows; row_index++ ) {
		const CalculationVector & row = CalculationVector( columns );
		mData.push_back( row );
	}		
}

Matrix::Matrix ( const Matrix & other ) :
	mData( other.mData ),
	mMapping( other.mMapping ) {}

void Matrix::insert ( const CalculationVector& row ) {
	mMapping.push_back( mData.size() );
	mData.push_back( row );
}

Matrix Matrix::operator * ( const Matrix & rhs ) const {
	Matrix result = *this;
	
	result *= rhs;

	return result;
}

void Matrix::operator *= ( const Matrix& matrix ) {

	CalculationVector buffer( matrix.getColumns() );

	for ( unsigned int row = 0; row < getRows(); row++ ) {
		
		for ( unsigned int column = 0; column < matrix.getColumns(); column++ ) {

			buffer.setValue( column, 0 );

			for ( unsigned int index = 0; index < getColumns(); index++ ) {
				if ( isNonZeroAt( row, index ) && matrix.isNonZeroAt( index, column ) ) {

					UtilsFactory::CALC_UTILS.loadRegisterT( getValue( row, index ) );
					UtilsFactory::CALC_UTILS.multiplyRegisterT( matrix.getValue( index, column ) );
					buffer.add( column, UtilsFactory::CALC_UTILS.getRegisterT() );

				}
			}
		}	
		getRow( row ) = buffer; 
	}
}

void Matrix::operator = ( const Matrix& matrix ) {
	mData = matrix.mData;
	mMapping = matrix.mMapping;
}

Matrix Matrix::getSubMatrix ( const unsigned int beginRow, const unsigned int endRow ) const {
	Matrix result;	

	for ( unsigned int index = beginRow; index < endRow; index++ ) {
		result.insert( getRow( index ) );
	}

	return result;
}

void Matrix::optimizeRowsOrder () {

	unsigned long min;

	for ( unsigned long i = getColumns(); i < getRows(); i++ ) {

		min = i;
		for ( unsigned long l = i + 1; l < getRows(); l++ ) {
			if ( getRow( min ).getNonZeroCount() >= getRow( l ).getNonZeroCount() ) {
				min = l;
			}
		}
		swapRows( min, i );
	} 
}


void Matrix::sort ( const SortCriteria & criteria, unsigned short start, unsigned short end, int swap_trigger ) {

	for ( unsigned short i = start; i < end; i++ ) {

		unsigned short pivot = i;
		for ( unsigned long l = i + 1; l < end; l++ ) {
			if ( criteria.cmp ( getRow( l ), getRow( pivot )) == swap_trigger ) {
				pivot = l;
			}
		}
		swapRows( pivot, i );
	} 	
}

const std::string Matrix::getMapping () const {
	std::stringstream output;

	output << "[";
	for ( unsigned long i = 0; i < getRows() - 1; i++ ) {
		output << mMapping[i] << ", ";
	}
	output << mMapping[ getRows()-1 ] << "]";

	return output.str();
}

const std::string Matrix::toString () const {
	std::stringstream output;

	for ( unsigned long i = 0; i < getRows() - 1; i++ ) {
		output << getRow( i ).toString() << std::endl;
	}
	output << getRow( getRows() - 1 ).toString();

	return output.str();
}

const std::string Matrix::toPortaFormat( const mpz_class & val, bool first = false ) const {
	std::stringstream output;
    
    if ( val > 0 && !first ) {
        output << "+";
    } else if ( val == -1 ) {
        output << "-";
    }
    if ( val != 1 && val != -1 ) {
        output << val;
    }

    return output.str();
}

const std::string Matrix::to4ti2Format () const {
    std::stringstream output;

    output << getRows() << " " << getColumns() << std::endl;

    for ( unsigned int r_index = 0; r_index < getRows(); r_index++ ) {
        const CalculationVector & row = getRow( r_index );

        for ( unsigned int c_index = 0; c_index < getColumns() - 1; c_index++ ) {
            output << row[ c_index ] << " ";
        }     
        output << row[ getColumns() - 1 ] << std::endl;
    }

    return output.str();
}

const std::string Matrix::to4ti2RelFormat () const {
    std::stringstream output;

    output << "1 " << getRows() << std::endl;

    for ( unsigned int r_index = 0; r_index < getRows() - 1; r_index++ ) {
        output << "> ";
    }
    output << ">" << std::endl;

    return output.str();
}

const std::string Matrix::to4ti2SignFormat () const {
    std::stringstream output;

    output << "1 " << getColumns() << std::endl;

    for ( unsigned int c_index = 0; c_index < getColumns() - 1; c_index++ ) {
        output << "0 ";
    }
    output << "0" << std::endl;

    return output.str();    
}

const std::string Matrix::toPortaFormat () const {
	std::stringstream output;

    output << "DIM = " << getColumns() << std::endl
           << "INEQUALITIES_SECTION" << std::endl;

    for ( unsigned int r_index = 0; r_index < getRows(); r_index++ ) {
        const CalculationVector & row = getRow( r_index );
        unsigned int c_index = 0;
        
        while ( !row.isNonZeroAt( c_index ) ) {
            c_index++;
        }

        output << toPortaFormat( row[ c_index ], true ) << "x" << ( c_index + 1 );
        c_index++;

        for ( ; c_index < getColumns(); c_index++ ) {
            if ( row.isNonZeroAt( c_index ) ) {
                output << toPortaFormat( row[ c_index ] ) << "x" << ( c_index + 1 );
            }
        }     
        output << ">=0" << std::endl;
    }

    output << "END";
    
    return output.str();
}


