/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_OPTIONS__
#define __DD_OPTIONS__

#include <string>
#include <sstream>
#include <fstream>
#include <algorithm> 
#include <functional> 
#include <cctype>
#include <locale>

#include "DDLogging.h"

#define SEC_GENERAL_SETTINGS                "GENERAL"
#define SEC_NARROWING_DEG_TREE              "NARROWING DEGENERATE TREE"
#define SEC_NARROWING_NON_DEG_TREE          "NARROWING NONDEGENERATE TREE"
#define SEC_VERIFICATION_TREE               "VERIFICATION TREE"

#define OPT_ADJACENCY_TEST                  "AdjacencyTest"
#define OPT_NARROWING_CONTAINER             "NarrowingContainer"
#define OPT_VERIFICATION_CONTAINER          "VerificationContainer"
#define OPT_PLOT_FILE                       "Plot file"
#define OPT_OUTPUT_FORMAT                   "Output format"

#define OPT_STRATEGY                        "Strategy"
#define OPT_MAX_TREE_DEPTH                  "MaxTreeDepth"
#define OPT_LEAF_CAPACITY                   "LeafCapacity"
#define OPT_DATA_SORT                       "DataSort"

#define OPT_BITS_PER_SPLIT                  "BitsPerSplit"
#define OPT_DYNAMIC_NARROWING_FACTOR        "DynamicNarrowingFactor"
#define OPT_DYNAMIC_BALANCE_FACTOR          "DynamicBalanceFactor"

#define OPT_VANTAGE_POINT_POPULATION        "VantagePointPopulation"
#define OPT_VANTAGE_POINT_DISTANCE_COEFF    "VantagePointDistanceCoefficient"
#define OPT_UNIQUE_VANTAGE_POINT            "UniqueVantagePoint"

namespace double_description {

    enum RayContainerType {
        STD_VECTOR,
        BIT_PATTERN_TREE,
        POPULATION_TREE,
        HYBRID_TREE
    };

    const std::string RayContainerTypeNames[] = {
        "STD_VECTOR",
        "BIT_PATTERN_TREE",
        "POPULATION_TREE",
        "HYBRID_TREE"
    };

    enum AdjacencyTest {
        COMBINATIONAL_TEST,
        ALGEBRAIC_TEST,
        PARTITIONED_SETS_TEST
    };

    const std::string AdjacencyTestNames[] = {
        "COMBINATIONAL_TEST",
        "ALGEBRAIC_TEST",
        "PARTITIONED_SETS_TEST"
    };

    const unsigned long AdjacencyTestNamesSize = 3;

    enum BPT_CreationStrategy {
        BPT_FIRST_AVAILABLE_BIT,
        BPT_MOST_COMMON_BIT_FIRST,
        BPT_LEAST_COMMON_BIT_FIRST,
        BPT_MOST_AVERAGE_BIT_FIRST,
        BPT_CLOSEST_TO_50_BIT_FIRST,
        BPT_QUERY_BITS_NEUTRALIZATION,
        BPT_LEAST_AVERAGE_BIT_FIRST,
        BINARY_DYNAMIC_VANTAGE_POINT,
        BINARY_STATIC_VANTAGE_POINT
    };

    const std::string BPT_CreationStrategyNames[] = {
        "BPT_FIRST_AVAILABLE_BIT",
        "BPT_MOST_COMMON_BIT_FIRST",
        "BPT_LEAST_COMMON_BIT_FIRST",
        "BPT_MOST_AVERAGE_BIT_FIRST",
        "BPT_CLOSEST_TO_50_BIT_FIRST",
        "BPT_QUERY_BITS_NEUTRALIZATION",
        "BPT_LEAST_AVERAGE_BIT_FIRST",
        "BINARY_DYNAMIC_VANTAGE_POINT",
        "BINARY_STATIC_VANTAGE_POINT"
    };

    const unsigned long CreationStrategyNamesSize = 9;

    enum OutputFormat {
        INTEGER_VEC,
        DOUBLE_VEC,
        ACTIVE_SET
    };

    const std::string OutputFormatNames[] = {
        "INTEGER_VEC",
        "DOUBLE_VEC",
        "ACTIVE_SET"
    };

    const unsigned long OutputFormatNamesSize = 3;

    enum SortType {
        MAX_SET_BITS,
        MIN_SET_BITS,
        MAX_SET_BITS_LEX,
        MIN_SET_BITS_LEX,
        MAX_BIT_FREQUENCY,
        MIN_BIT_FREQUENCY,
        MAX_SET_BITS_MAX_BIT_FREQUENCY,
        MAX_SET_BITS_MIN_BIT_FREQUENCY,
        MIN_SET_BITS_MAX_BIT_FREQUENCY,
        MIN_SET_BITS_MIN_BIT_FREQUENCY,
        NONE
    };

    const std::string SortTypeNames[] = {
        "MAX_SET_BITS",
        "MIN_SET_BITS",
        "MAX_SET_BITS_LEX",
        "MIN_SET_BITS_LEX",
        "MAX_BIT_FREQUENCY",
        "MIN_BIT_FREQUENCY",
        "MAX_SET_BITS_MAX_BIT_FREQUENCY",
        "MAX_SET_BITS_MIN_BIT_FREQUENCY",
        "MIN_SET_BITS_MAX_BIT_FREQUENCY",
        "MIN_SET_BITS_MIN_BIT_FREQUENCY",
        "NONE"
    };

    const unsigned int SortTypeNamesSize = 17;

    class TreeOptions {

        private:
        const std::string mCaption;

        BPT_CreationStrategy mStrategy;

        unsigned long mMaxTreeDepth;
        unsigned long mLeafCapacity;

        SortType mDataSort;

        unsigned long mBitsPerSplit;
        double mDynamicNarrowingFactor;
        double mDynamicBalanceFactor;

        unsigned long mVantagePointPopulation;
        double mVantagePointDistanceCoefficient;  
        bool mUniqueVantagePoint;

        public:
        TreeOptions ( const std::string caption ) :
            mCaption( caption ),
            mStrategy( BPT_FIRST_AVAILABLE_BIT ),
            mMaxTreeDepth( 15 ),
            mLeafCapacity( 20 ),
            mDataSort( NONE ),
            mBitsPerSplit( 1 ),
            mDynamicNarrowingFactor( 1.5 ),
            mDynamicBalanceFactor( 0.0 ),
            mVantagePointPopulation( 50 ),
            mVantagePointDistanceCoefficient( 0.0 ),
            mUniqueVantagePoint( false ) {}  

        void set ( const std::string & name, const std::string & value );

        BPT_CreationStrategy strategy () const {
            return mStrategy;
        }

        unsigned long max_tree_depth () const {
            return mMaxTreeDepth;
        }

        unsigned long leaf_capacity () const {
            return mLeafCapacity;
        }

        unsigned short bits_per_split () const {
            return mBitsPerSplit;
        }

        unsigned short vantage_point_population () const {
            return mVantagePointPopulation;
        }

        double vantage_point_distance_coefficient () const {
            return mVantagePointDistanceCoefficient;
        }

        double dynamic_narrowing_factor () const {
            return mDynamicNarrowingFactor;
        }

        double dynamic_balance_factor () const {
            return mDynamicBalanceFactor;
        }

        SortType data_sort () const {
            return mDataSort;
        }         

        bool unique_vp () const {
            return mUniqueVantagePoint;
        }

        const std::string toString () const;
    };

    class Options {
        
        private:
        AdjacencyTest mAdjacencyTest;
        RayContainerType mPreelimContainer;
        RayContainerType mVerContainer;
        OutputFormat mOutputFormat;

        std::string mPlotFile;

        TreeOptions mDegNarrowingTree;
        TreeOptions mNonDegNarrowingTree;
        TreeOptions mVerificationTree;

        static const char COMMENTS_DELIM;
        static const char ASSIGNMENT_DELIM;

        Options () :
            mAdjacencyTest( COMBINATIONAL_TEST ),
            mPreelimContainer( BIT_PATTERN_TREE ),
            mVerContainer( BIT_PATTERN_TREE ),
            mOutputFormat( INTEGER_VEC ),
            mPlotFile( "Don't plot" ),
            mDegNarrowingTree( SEC_NARROWING_DEG_TREE ),
            mNonDegNarrowingTree( SEC_NARROWING_NON_DEG_TREE ),
            mVerificationTree( SEC_VERIFICATION_TREE ) {}

        static Options * mCurrent;

        void set ( const std::string & name, const std::string & value );

        public:
        static Options & current () {

            if ( mCurrent == NULL ) {
                mCurrent = new Options();
            } 
            
            return *mCurrent;
        }        

        static void release () {
            if ( mCurrent != NULL ) {
                delete mCurrent;
            } 
        }

        static void trim ( std::string & s ) {
            s.erase( s.begin(), std::find_if(s.begin(), s.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
            s.erase( std::find_if(s.rbegin(), s.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());
        }

        static void remove_comments ( std::string & s, const char delim ) {
            size_t p = s.find( delim );

            if ( p != std::string::npos ) {
                s.erase( s.begin() + p, s.end() );
            }
        }

        bool readSettingsFile ( const char * file );

        static void write_value ( const std::string & s_value, bool & b_value ) {
            std::istringstream is_value( s_value );
            is_value >> b_value;        
        }

        static void write_value ( const std::string & s_value, unsigned long & ul_value ) {
            std::istringstream is_value( s_value );
            is_value >> ul_value;        
        }

        static void write_value ( const std::string & s_value, double & d_value ) {
            std::istringstream is_value( s_value );
            is_value >> d_value;        
        }

        template< typename EnumType >
        static void write_value ( const std::string & s_value,
                           EnumType & e_value,
                           const std::string* names,
                           const unsigned int size ) {
            const std::string * res = std::find( names, names + size, s_value );
            if ( res != names + size ) {
                e_value = static_cast<EnumType>( res - names );
            }   
        }

        OutputFormat output_format () const {
            return mOutputFormat;
        }

        AdjacencyTest adjacency_test () const {
            return mAdjacencyTest;
        }

        RayContainerType narrowing_container_type () const {
            return mPreelimContainer;
        }

        RayContainerType verification_container_type () const {
            return mVerContainer;
        }
   
        const TreeOptions & deg_narrowing () const {
            return mDegNarrowingTree;
        }

        const TreeOptions & non_deg_narrowing () const {
            return mNonDegNarrowingTree;
        }        

        const TreeOptions & verification () const {
            return mVerificationTree;
        }

        const std::string & plot_file () const {
            return mPlotFile;
        }

        bool plot_required () const {
            return mPlotFile != "Don't plot";
        }
    
        const std::string getCaptionBlock ( const std::string text ) const;
        const std::string toString () const;
    };
}

#endif // __DD_OPTIONS__
