
#ifndef __DD_POPULATION_TREE_NODE__
#define __DD_POPULATION_TREE_NODE__

#include <vector>

#include "DDBitList.h"
#include "DDTreeRayContainer.h"

#include "DDBitSetTreeNode.h"

namespace double_description {

	template <class RAY_TYPE>
	class PopulationTreeNode;

	template <class RAY_TYPE>
	class SplitCriteria {

		private:
		const unsigned short mMaxDepth;
		const unsigned short mMaxSplits;

		public:
		SplitCriteria( const unsigned short maxDepth, const unsigned short maxSplits ) :
			mMaxDepth( maxDepth ),
			mMaxSplits( maxSplits ) {}

		bool operator() ( const PopulationTreeNode<RAY_TYPE>* node, const unsigned short depth ) const {
		
			const BitList & positive_hash = node->getPositiveHash();

			return ( mMaxDepth > depth && positive_hash.getCapacity() - positive_hash.getPopulationCount() < mMaxSplits );
		}
	};

	template <class RAY_TYPE>
	class PopulationTreeNode {

		private:
		BitList mPattern;
		BitList mPositiveHash;
		BitList mNegativeHash;

		PopulationTreeNode* mPatternUnder;
		PopulationTreeNode* mInverseUnder;
		PopulationTreeNode* mDoubleOver;

		TreeRayContainer<RAY_TYPE> * mContainer;

		BitSetTreeNode<RAY_TYPE> * mBitPatternTree;

		BitList mMap;
		std::vector<const RAY_TYPE*> mRays;

		void copyRays ( const unsigned short patternThreshold,
						const unsigned short inverseThreshold ) {

			for ( unsigned long index = 0; index < mRays.size(); index++ ) {

				const unsigned int pattern_popcnt = mPattern.getCommonPopCnt( mRays[index]->getActiveSet() );

				if ( pattern_popcnt < patternThreshold ) {
					mPatternUnder->insert( mRays[index], patternThreshold, inverseThreshold );
				} else {
					const unsigned int inverse_popcnt = mPattern.getInvCommonPopCnt( mRays[index]->getActiveSet() );

					if ( inverse_popcnt < inverseThreshold ) {
						mInverseUnder->insert( mRays[index], patternThreshold, inverseThreshold );
					} else {
						mDoubleOver->insert( mRays[index], patternThreshold, inverseThreshold );
					}
				}				
			}
			mRays.clear();

		/*	const std::vector<const RAY_TYPE*> positive = mContainer->getPositiveRays();
			const std::vector<const RAY_TYPE*> negative = mContainer->getNegativeRays();
			const std::vector<const RAY_TYPE*> zero = mContainer->getZeroRays();

			for ( unsigned long index = 0; index < positive.size(); index++ ) {

				const unsigned int pattern_popcnt = mPattern.getCommonPopCnt( positive[index]->getActiveSet() );

				if ( pattern_popcnt < patternThreshold ) {
					mPatternUnder->getContainer()->insertPositive( positive[index] );
				} else {
					const unsigned int inverse_popcnt = mPattern.getInvCommonPopCnt( positive[index]->getActiveSet() );

					if ( inverse_popcnt < inverseThreshold ) {
						mInverseUnder->getContainer()->insertPositive( positive[index] );
					} else {
						mDoubleOver->getContainer()->insertPositive( positive[index] );
					}
				}				
			}

			for ( unsigned long index = 0; index < negative.size(); index++ ) {

				const unsigned int pattern_popcnt = mPattern.getCommonPopCnt( negative[index]->getActiveSet() );

				if ( pattern_popcnt < patternThreshold ) {
					mPatternUnder->getContainer()->insertNegative( negative[index] );
				} else {
					const unsigned int inverse_popcnt = mPattern.getInvCommonPopCnt( negative[index]->getActiveSet() );

					if ( inverse_popcnt < inverseThreshold ) {
						mInverseUnder->getContainer()->insertNegative( negative[index]);
					} else {
						mDoubleOver->getContainer()->insertNegative( negative[index] );
					}
				}				
			}

			for ( unsigned long index = 0; index < zero.size(); index++ ) {

				const unsigned int pattern_popcnt = mPattern.getCommonPopCnt( zero[index]->getActiveSet() );

				if ( pattern_popcnt < patternThreshold ) {
					mPatternUnder->getContainer()->insertZero( zero[index] );
				} else {
					const unsigned int inverse_popcnt = mPattern.getInvCommonPopCnt( zero[index]->getActiveSet() );

					if ( inverse_popcnt < inverseThreshold ) {
						mInverseUnder->getContainer()->insertZero( zero[index] );
					} else {
						mDoubleOver->getContainer()->insertZero( zero[index] );
					}
				}				
			} */
		}

		public:
		PopulationTreeNode( const BitList & positive, const BitList & negative, const BitList & pattern ) :
			mPattern( pattern ),
			mPositiveHash( positive ),
			mNegativeHash( negative ),
			mPatternUnder( NULL ),
			mInverseUnder( NULL ),
			mDoubleOver( NULL ),
			mBitPatternTree( NULL ),
			mMap( 2 ) {} // @ToDo Fix this!

		PopulationTreeNode( const PopulationTreeNode<RAY_TYPE> & other ) :
			mPattern( other.mPattern ),
			mPositiveHash( other.mPositiveHash ),
			mNegativeHash( other.mNegativeHash ),
			mPatternUnder( other.mPatternUnder ),
			mInverseUnder( other.mInverseUnder ),
			mDoubleOver( other.mDoubleOver ),
			mBitPatternTree( other.mBitPatternTree ),
			mMap( other.mMap ),
			mRays( other.mRays ) {}

		void partition( const SplitCriteria<RAY_TYPE> & criteria, std::vector<const PopulationTreeNode<RAY_TYPE>*> & leafs, const unsigned short blocks,
			const BitList * const patternSet, const unsigned short patternID = 0 ) {
			
			if ( criteria( this, patternID + 1 ) ) {
				BitList pattern_under_pos_hash = mPositiveHash;
				BitList pattern_under_neg_hash = mNegativeHash;
				BitList inverse_under_pos_hash = mPositiveHash;
				BitList inverse_under_neg_hash = mNegativeHash;

				pattern_under_pos_hash.unsetBit( 2 * patternID );
				pattern_under_neg_hash.unsetBit( 2 * patternID + 1 );
				inverse_under_pos_hash.unsetBit( 2 * patternID + 1 );
				inverse_under_neg_hash.unsetBit( 2 * patternID );

				mPatternUnder = new PopulationTreeNode( pattern_under_pos_hash, pattern_under_neg_hash, patternSet[patternID + 1] );
				mInverseUnder = new PopulationTreeNode( inverse_under_pos_hash, inverse_under_neg_hash, patternSet[patternID + 1] );
				mDoubleOver = new PopulationTreeNode( mPositiveHash, mNegativeHash, patternSet[patternID + 1] );

				mPatternUnder->partition( criteria, leafs, blocks, patternSet, patternID + 1 );
				mInverseUnder->partition( criteria, leafs, blocks, patternSet, patternID + 1 );
				mDoubleOver->partition( criteria, leafs, blocks, patternSet, patternID + 1 );
			} else {
				mContainer = new TreeRayContainer<RAY_TYPE>( blocks );
				leafs.push_back( this );
			}
		}

		void partition( const unsigned short blocks,
						const BitList * const patternSet, const unsigned short patternThreshold,
						const unsigned short inverseThreshold, const unsigned short patternID = 0 ) {
			
			if ( mRays.size() > 16 && patternID < 15 ) {

					BitList pattern_under_pos_hash = mPositiveHash;
					BitList pattern_under_neg_hash = mNegativeHash;
					BitList inverse_under_pos_hash = mPositiveHash;
					BitList inverse_under_neg_hash = mNegativeHash;

					pattern_under_pos_hash.unsetBit( 2 * patternID );
					pattern_under_neg_hash.unsetBit( 2 * patternID + 1 );
					inverse_under_pos_hash.unsetBit( 2 * patternID + 1 );
					inverse_under_neg_hash.unsetBit( 2 * patternID );

					mPatternUnder = new PopulationTreeNode( pattern_under_pos_hash, pattern_under_neg_hash, patternSet[patternID + 1] );
					mInverseUnder = new PopulationTreeNode( inverse_under_pos_hash, inverse_under_neg_hash, patternSet[patternID + 1] );
					mDoubleOver = new PopulationTreeNode( mPositiveHash, mNegativeHash, patternSet[patternID + 1] );

					copyRays( patternThreshold, inverseThreshold );

				/*	delete mContainer;
					mContainer = NULL; */

					mPatternUnder->partition( blocks, patternSet, patternThreshold, inverseThreshold, patternID + 1 );
					mInverseUnder->partition( blocks, patternSet, patternThreshold, inverseThreshold, patternID + 1 );
					mDoubleOver->partition( blocks, patternSet, patternThreshold, inverseThreshold, patternID + 1 );
			} else {

				mBitPatternTree = new BitSetTreeNode<RAY_TYPE>( blocks );

				for ( unsigned long index = 0; index < mRays.size(); index++ ) {
					mBitPatternTree->insert( mRays[index] );
				}

				mBitPatternTree->partition( 14 );

				mRays.clear();
			}
		}

		const BitSetTreeNode<RAY_TYPE> * getBitPatternTree () const {
			return mBitPatternTree;
		}

		bool isLeaf () const {
			return mDoubleOver == NULL;
		}

		const BitList & getPositiveHash () const {
			return mPositiveHash;
		}

		const BitList & getNegativeHash () const {
			return mNegativeHash;
		}

		bool areCombinable ( const PopulationTreeNode<RAY_TYPE> * const other ) const {
			return mPositiveHash.isUnionFull( other->mNegativeHash );
		}

		const BitList & getPattern () const {
			return mPattern;
		}

		const BitList & getMap () const {
			return mMap;
		}

		const std::vector<const RAY_TYPE*> & getRays() const {
			return mRays;
		}

		PopulationTreeNode * getPatternUnderNode () const {
			return mPatternUnder;
		}

		PopulationTreeNode * getInverseUnderNode () const {
			return mInverseUnder;
		}
		
		PopulationTreeNode * getDoubleOverNode () const {
			return mDoubleOver;
		}

		const TreeRayContainer<SingleSetRay> * getContainer () const {
			return mContainer;
		}

		TreeRayContainer<SingleSetRay> * getContainer () {
			return mContainer;
		}

		void insert ( const RAY_TYPE * const ray, /*const RayPosition pos,*/ const unsigned short patternThreshold, const unsigned short inverseThreshold ) {
			
			if ( isLeaf() ) {
				mMap |= ray->getActiveSet();
				mRays.push_back( ray );

			//	mContainer->insert( ray, pos );
			} else {
				const unsigned int pattern_popcnt = mPattern.getCommonPopCnt( ray->getActiveSet() );

				if ( pattern_popcnt < patternThreshold ) {
					mPatternUnder->insert( ray, /*pos,*/ patternThreshold, inverseThreshold );
				} else {
					const unsigned int inverse_popcnt = mPattern.getInvCommonPopCnt( ray->getActiveSet() );

					if ( inverse_popcnt < inverseThreshold ) {
						mInverseUnder->insert( ray, /*pos,*/ patternThreshold, inverseThreshold );
					} else {
						mDoubleOver->insert( ray, /*pos,*/ patternThreshold, inverseThreshold );
					}
				}
			}
		}
	};
}

#endif // __DD_BIT_SET_TREE_NODE__
