/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_RAY__
#define __DD_RAY__

#include "DDBitList.h"
#include "DDVector.h"
#include "DDRepresentationVector.h"
#include "DDCalculationVector.h"
#include "DDUtilsFactory.h"

#include <string.h>

#include <gmp.h>
#include <gmpxx.h>

#include "DDLogging.h"

namespace double_description {

#ifdef __COMPRESSION__
	class Ray {
#else
	class Ray : public RepresentationVector {
#endif // __COMPRESSION__

		private:
#ifdef __COMPRESSION__
		const unsigned short dim;
#endif // __COMPRESSION__
		const unsigned short mCreationPoint;
		mpz_class* mPlaneCoeff;

		public:
		Ray (	const unsigned int dimension, mpz_class* const coeffs ) :
#ifdef __COMPRESSION__
			dim( dimension ),
#else
			RepresentationVector( dimension ),
#endif // __COMPRESSION__
			mCreationPoint( dimension - 1 ),
			mPlaneCoeff( coeffs ) {}

		Ray (	const unsigned int dimension, const unsigned int creationPoint, const unsigned int coeffSize ) :
#ifdef __COMPRESSION__
			dim( dimension ),
#else
			RepresentationVector( dimension ),
#endif // __COMPRESSION__
			mCreationPoint( creationPoint ) {
	
			mPlaneCoeff = new mpz_class[coeffSize];
		}

		Ray ( const RepresentationVector & vector, const unsigned int creationPoint, mpz_class* const coeffs ) :
#ifdef __COMPRESSION__
			dim( vector.getDimension() ),
#else
			RepresentationVector( vector ),
#endif // __COMPRESSION__
			mCreationPoint( creationPoint ),
			mPlaneCoeff( coeffs ) {}

		Ray ( const Ray* const positive, const Ray* const negative, const unsigned int plane, const unsigned int totalPlanes ) :
#ifdef __COMPRESSION__
			dim( positive->getDimension() ),
#else
			RepresentationVector( positive->getDimension() ),
#endif // __COMPRESSION__
			mCreationPoint( plane ) {

			UtilsFactory::CALC_UTILS.calcGCD( positive->getCachedPlaneCoeff( plane ), negative->getCachedPlaneCoeff( plane ) );

			UtilsFactory::CALC_UTILS.loadRegisterP( positive->getCachedPlaneCoeff( plane ) );
			UtilsFactory::CALC_UTILS.divideRegisterP( UtilsFactory::CALC_UTILS.getResult() );

			UtilsFactory::CALC_UTILS.loadRegisterQ( negative->getCachedPlaneCoeff( plane ) );
			UtilsFactory::CALC_UTILS.divideRegisterQ( UtilsFactory::CALC_UTILS.getResult() );

#ifndef __COMPRESSION__
			for ( unsigned int index = 0; index < getDimension(); index++ ) {
				if ( positive->isNonZeroAt( index ) ) {

					UtilsFactory::CALC_UTILS.loadRegisterS( positive->getValue( index ) );

					if ( negative->isNonZeroAt( index ) ) {
						UtilsFactory::CALC_UTILS.loadRegisterT( negative->getValue( index ) ); 
						UtilsFactory::CALC_UTILS.multiplyRegisterT( UtilsFactory::CALC_UTILS.getRegisterP() );
	
						UtilsFactory::CALC_UTILS.multiplyRegisterS( UtilsFactory::CALC_UTILS.getRegisterQ() );
						UtilsFactory::CALC_UTILS.subtractRegisterT( UtilsFactory::CALC_UTILS.getRegisterS() );
						setValue( index, UtilsFactory::CALC_UTILS.getRegisterT() );
					} else {
						UtilsFactory::CALC_UTILS.multiplyRegisterS( UtilsFactory::CALC_UTILS.getRegisterQ() );
						UtilsFactory::CALC_UTILS.negateRegisterS();
						setValue( index, UtilsFactory::CALC_UTILS.getRegisterS() );
					}
			
				} else if ( negative->isNonZeroAt( index ) ) {
					UtilsFactory::CALC_UTILS.loadRegisterT( negative->getValue( index ) );
					UtilsFactory::CALC_UTILS.multiplyRegisterT( UtilsFactory::CALC_UTILS.getRegisterP() );
					setValue( index, UtilsFactory::CALC_UTILS.getRegisterT() );
				} 
			}
#endif // __COMPRESSION__

			mPlaneCoeff = new mpz_class[ totalPlanes - ( plane + 1) ];

			for ( unsigned int next_plane = plane + 1; next_plane < totalPlanes; next_plane++ ) {

				UtilsFactory::CALC_UTILS.loadRegisterS( UtilsFactory::CALC_UTILS.getRegisterP() );
				UtilsFactory::CALC_UTILS.multiplyRegisterS( negative->getCachedPlaneCoeff( next_plane ) );

				UtilsFactory::CALC_UTILS.loadRegisterT( UtilsFactory::CALC_UTILS.getRegisterQ() );
				UtilsFactory::CALC_UTILS.multiplyRegisterT( positive->getCachedPlaneCoeff( next_plane ) );	

				UtilsFactory::CALC_UTILS.subtractRegisterS( UtilsFactory::CALC_UTILS.getRegisterT() );	

				mPlaneCoeff[ next_plane - ( plane + 1) ] = UtilsFactory::CALC_UTILS.getRegisterS();
			}
		}

		virtual ~Ray () {
			delete[] mPlaneCoeff;
		}

#ifdef __COMPRESSION__
		unsigned short getDimension () const {
			return dim;
		}
#endif

		const mpz_class & getCachedPlaneCoeff ( const unsigned int plane ) const {
			return mPlaneCoeff[ plane - ( mCreationPoint + 1 ) ];
		}

		mpz_class & getCachedPlaneCoeff ( const unsigned int plane ) {
			return mPlaneCoeff[ plane - ( mCreationPoint + 1 ) ];
		}

		void setCachedPlaneCoeff ( const unsigned int plane, const mpz_class & value ) {
			mPlaneCoeff[ plane - ( mCreationPoint + 1 ) ] = value;
		}

		bool isInside ( const unsigned int plane ) const {
			return getCachedPlaneCoeff ( plane ) > 0;
		}
		
		bool isOutside ( const unsigned int plane ) const {
			return getCachedPlaneCoeff ( plane ) < 0;
		}
		
		bool isOnThePlane ( const unsigned int plane ) const {
			return getCachedPlaneCoeff ( plane ) == 0;
		}

		bool isInitial () const {
			return getCreationPoint() == getDimension() - 1;
		}

		unsigned short getCreationPoint () const {
			return mCreationPoint;
		}

		const std::string toString () const;

	};
}

#endif // __DD_RAY__
