
#ifndef __DD_RAY_CACHE_ENTRY__
#define __DD_RAY_CACHE_ENTRY__

#include <vector>
#include <sstream>

#include "DDStats.h"

#include "DDBitList.h"
#include "DDBitCountTable.h"
#include "DDSingleSetRay.h"
#include "DDDualSetRay.h"
#include "DDRayPatternTree.h"
#include "DDPatternGraph.h"

namespace double_description {
	
	template <class RAY_TYPE>
	class RayCacheEntry {

		private:
		unsigned short mBlocks;
		RayPatternTree<RAY_TYPE> positive;
		RayPatternTree<RAY_TYPE> negative;

		unsigned long* mPositiveMasks;
		unsigned long* mNegativeMasks;

		static PatternGraph sGraph;

		static const unsigned long POSITIVE_PATTERNS[]; 
		static const unsigned long NEGATIVE_PATTERNS[];

		public:
		RayCacheEntry ( unsigned short plane, unsigned char threshold, unsigned char depth = 14 ) :
			mBlocks( BitList::getBlocks( plane + 1 ) ),
			positive( 14, mBlocks, threshold ),
			negative( 14, mBlocks, threshold - 1 ) {

			mPositiveMasks = new unsigned long[ depth * mBlocks ];
			mNegativeMasks = new unsigned long[ depth * mBlocks ];

			for ( unsigned char level = 0; level < depth; level++ ) {			
				for ( unsigned short block = 0; block < mBlocks; block++ ) {
					mPositiveMasks[level*mBlocks + block] = POSITIVE_PATTERNS[level];
					if ( block + 1 == mBlocks ) {
						mPositiveMasks[level*mBlocks + block] &= ( __SINGLE_BIT__ << ( plane % __BLOCK_SIZE__ ) ) - 1;
					}
					mNegativeMasks[level*mBlocks + block] = NEGATIVE_PATTERNS[level];
				}
			}

			positive.setPatterns( mPositiveMasks );
			negative.setPatterns( mNegativeMasks );
		}

		~RayCacheEntry () {}
		
		void insertPositive ( const RAY_TYPE* const ray ) {

			positive.insert( ray );
		}

		void insertNegative ( const RAY_TYPE* const ray ) {

			negative.insert( ray );
		}

        void insertZero ( const RAY_TYPE* const ) {
			/* No action needed here. Should be fixed if dynamic caching is to be used! */
		}

		const struct ActiveSetContainer<RAY_TYPE> getPositiveContainer () const {
			return positive.getActiveSetContainer();
		}

		const struct ActiveSetContainer<RAY_TYPE> getNegativeContainer () const {
			return negative.getActiveSetContainer();
		}

		unsigned long getPositiveCount() const {
			return positive.getRayCount();
		}

		unsigned long getNegativeCount() const {
			return negative.getRayCount();
		}

		void clear() {
			positive.clear();
			negative.clear();

			delete[] mPositiveMasks;
			delete[] mNegativeMasks;
		}

		const struct StepStats fillCandidatesBuffer ( std::vector<const RAY_TYPE*> & buffer, const unsigned short minCommonBits ) const {

            struct StepStats stats;

			unsigned long tests = 0;

			const struct ActiveSetContainer<SingleSetRay> p_container = getPositiveContainer();
			const struct ActiveSetContainer<SingleSetRay> n_container = getNegativeContainer();

			for ( unsigned short p_cell = 0; p_cell < p_container.cells; p_cell++ ) {
				const unsigned long* p_end_point = p_container.starting_points[p_cell+1];

				for ( const unsigned long* positive_index = p_container.starting_points[p_cell]; positive_index < p_end_point; positive_index += __builtin_expect ( mBlocks, 1 ) ) {

					const unsigned short* edges = sGraph.getEdges( p_cell );
					const unsigned short edges_count = sGraph.getEdgesCount( p_cell );

					for ( unsigned short n_cell = 0; n_cell < edges_count; n_cell++ ) {

						if ( !negative.getUnionMap( edges[n_cell] ).isEmpty() &&
							p_container.rays[ ( positive_index - p_container.sets )/ mBlocks ]->getActiveSet().getCommonPopCnt( negative.getUnionMap( edges[n_cell] )) >= minCommonBits ) {

						const unsigned long* n_end_point = n_container.starting_points[edges[n_cell]+1];

						tests += ( n_end_point - n_container.starting_points[edges[n_cell]] ) / mBlocks;

						for ( const unsigned long* negative_index = n_container.starting_points[edges[n_cell]]; negative_index < n_end_point; negative_index += __builtin_expect ( mBlocks, 1 ) ) {

							register const unsigned long common_bits = BitList::BIT_COUNT_TABLE.lookUp( positive_index, negative_index, mBlocks );

							if ( __builtin_expect ( common_bits >= minCommonBits, 0 ) ) {
								buffer.push_back( p_container.rays[ ( positive_index - p_container.sets )/__builtin_expect ( mBlocks, 1 )] );
								buffer.push_back( n_container.rays[ ( negative_index - n_container.sets )/__builtin_expect ( mBlocks, 1 )] ); 
							}
						}
						}
					}
				}
			}

		// @ToDo Fix this. The rays can't be freed here!
		/*	for ( unsigned long index = 0; index < n_container.ray_count; index++ ) {
				delete n_container.rays[index];
			} */

			delete[] n_container.rays;
			delete[] n_container.sets;
			delete[] n_container.starting_points;

			delete[] p_container.rays;
			delete[] p_container.sets;
			delete[] p_container.starting_points;

			return stats;
		}

		const std::string toString () const {
			std::stringstream output;

			output << positive.getSize() << "positive, " << negative.getSize() << " negative" << std::endl;

			return output.str();
		}
	};

	template <class RAY_TYPE>
	PatternGraph RayCacheEntry<RAY_TYPE>::sGraph( 14, 6 );

#if defined _LP64 || defined __LP64__
	template <class RAY_TYPE>
	const unsigned long RayCacheEntry<RAY_TYPE>::POSITIVE_PATTERNS[] = {
		0x5555555555555555,
		0xAAAAAAAAAAAAAAAA,
		0x9999999999999999,
		0x6666666666666666,
		0xF0F0F0F0F0F0F0F0,
		0x0F0F0F0F0F0F0F0F,
		0x3333333333333333, 
		0xCCCCCCCCCCCCCCCC, 
		0xC3C3C3C3C3C3C3C3,
		0x3C3C3C3C3C3C3C3C,
		0x6969696969696969,
		0x9696969696969696,
		0xA5A5A5A5A5A5A5A5,
		0x5A5A5A5A5A5A5A5A,
		0x0000FFFF0000FFFF,
		0xFFFF0000FFFF0000,
		0x00FF00FF00FF00FF,
		0xFF00FF00FF00FF00
	};

	template <class RAY_TYPE>
	const unsigned long RayCacheEntry<RAY_TYPE>::NEGATIVE_PATTERNS[] = {
		0xAAAAAAAAAAAAAAAA,
		0x5555555555555555,
		0x6666666666666666,
		0x9999999999999999,
		0x0F0F0F0F0F0F0F0F,
		0xF0F0F0F0F0F0F0F0,
		0xCCCCCCCCCCCCCCCC, 
		0x3333333333333333, 
		0x3C3C3C3C3C3C3C3C,
		0xC3C3C3C3C3C3C3C3,
		0x9696969696969696,
		0x6969696969696969,
		0x5A5A5A5A5A5A5A5A,
		0xA5A5A5A5A5A5A5A5,
		0xFFFF0000FFFF0000, 
		0x0000FFFF0000FFFF,
		0xFF00FF00FF00FF00,
		0x00FF00FF00FF00FF
	}; 
#else
	template <class RAY_TYPE>
	const unsigned long RayCacheEntry<RAY_TYPE>::POSITIVE_PATTERNS[] = {
		0x55555555,
		0xAAAAAAAA,
		0x99999999,
		0x66666666,
		0xF0F0F0F0,
		0x0F0F0F0F,
		0x33333333, 
		0xCCCCCCCC, 
		0xC3C3C3C3,
		0x3C3C3C3C,
		0x69696969,
		0x96969696,
		0xA5A5A5A5,
		0x5A5A5A5A,
		0x0000FFFF,
		0xFFFF0000,
		0x00FF00FF,
		0xFF00FF00
	};

	template <class RAY_TYPE>
	const unsigned long RayCacheEntry<RAY_TYPE>::NEGATIVE_PATTERNS[] = {
		0xAAAAAAAA,
		0x55555555,
		0x66666666,
		0x99999999,
		0x0F0F0F0F,
		0xF0F0F0F0,
		0xCCCCCCCC, 
		0x33333333, 
		0x3C3C3C3C,
		0xC3C3C3C3,
		0x96969696,
		0x69696969,
		0x5A5A5A5A,
		0xA5A5A5A5,
		0xFFFF0000, 
		0x0000FFFF,
		0xFF00FF00,
		0x00FF00FF
	}; 
#endif // _LP64 || defined __LP64__
}

#endif // __DD_RAY_CACHE_ENTRY__
