
#ifndef __DD_RAY_CACHE_PATTERN_TREE__
#define __DD_RAY_CACHE_PATTERN_TREE__

#include <vector>

#include "DDBitList.h"
#include "DDSingleSetRay.h"
#include "DDDualSetRay.h"

#include "DDLogging.h"

namespace double_description {

	template <class RAY_TYPE>
	struct ActiveSetContainer {
		unsigned short cells;
		unsigned long ray_count;
		unsigned long * sets;
		const RAY_TYPE** rays;
		unsigned long ** starting_points;
	};

	template <class RAY_TYPE>
	class RayPatternTree {
		
		private:
		unsigned char mDepth;
		unsigned short mCells;
		unsigned char mThreshold;
		unsigned long mRayCount;
		const unsigned long* mPatterns;
		unsigned short mBlocks;
		std::vector<const RAY_TYPE*>* mLeafNodes;
		std::vector<BitList> mBitUnions;

		unsigned long* mBuffer;
		unsigned long* mStats;

		public:
		RayPatternTree ( unsigned char depth, const unsigned long * patterns, const unsigned short blocks, const unsigned char threshold ) :
			mDepth( depth ),
			mCells( 1 << depth ),
			mThreshold( threshold ),
			mRayCount( 0 ),
			mPatterns( patterns ),
			mBlocks( blocks ),
			mBitUnions( mCells, BitList( blocks ) ) {

				mLeafNodes = new std::vector<const RAY_TYPE*>[mCells];
				mBuffer = new unsigned long[blocks];
				mStats = new unsigned long[mDepth];

				for ( unsigned char i = 0; i < mDepth; i++ ) {
					mStats[i] = 0;
				}
		}

		RayPatternTree ( unsigned char depth, const unsigned short blocks, const unsigned char threshold ) :
			mDepth( depth ),
			mCells( 1 << depth ),
			mThreshold( threshold ),
			mRayCount( 0 ),
			mPatterns( NULL ),
			mBlocks( blocks ),
			mBitUnions( mCells, BitList( blocks ) ) {

				mLeafNodes = new std::vector<const RAY_TYPE*>[mCells];
				mBuffer = new unsigned long[blocks];
				mStats = new unsigned long[mDepth];
		}

		~RayPatternTree () {}

		void setPatterns ( const unsigned long* patterns ) {
			mPatterns = patterns;
		}

		void insert ( const RAY_TYPE* const ray ) {
			unsigned short hash = 0;
			unsigned char unders = 0;

			ray->copyActiveSetData( mBuffer, mBlocks );

			for ( unsigned short p = 0; p < mDepth; p++ ) {
				if ( unders > 6 || BitList::BIT_COUNT_TABLE.lookUp ( mBuffer, mPatterns + p*mBlocks, mBlocks ) > mThreshold ) {
					hash |= ( 1 << p );
				} else {
					unders++;
				}
			}

			mLeafNodes[hash].push_back( ray );
			mBitUnions[hash] |= ray->getActiveSet();
			mRayCount++;
		}

		const struct ActiveSetContainer<RAY_TYPE> getActiveSetContainer () const {
			
			unsigned long g_index = 0;

			struct ActiveSetContainer<RAY_TYPE> container;
			container.cells = mCells;
			container.ray_count = mRayCount;
			container.sets = new unsigned long[ mRayCount * mBlocks ];
			container.rays = new const RAY_TYPE*[ mRayCount ];
			container.starting_points = new unsigned long*[mCells + 1];

			for ( unsigned short cell = 0; __builtin_expect( cell < mCells, 1 ); cell++ ) {
				container.starting_points[cell] = container.sets + g_index * mBlocks;
				for ( unsigned long cell_index = 0; cell_index < mLeafNodes[cell].size(); cell_index++ ) {
					container.rays[g_index] = mLeafNodes[cell][cell_index];
					mLeafNodes[cell][cell_index]->copyActiveSetData( container.sets + g_index++ * mBlocks, mBlocks );
				}	
			}
			
			container.starting_points[ mCells ] = container.sets + mRayCount * mBlocks;		

			return container;
		}

		unsigned long getRayCount() const {
			return mRayCount;
		}

		const BitList & getUnionMap( const unsigned short cell ) const {
			return mBitUnions[ cell ];
		}

		unsigned long getSize() const {
			unsigned long size = 0;

			for ( unsigned short i = 0; i < mCells; i++ ) {
				size += mLeafNodes[i].size();
			}

			return size;
		}

		void clear () {
			if ( mLeafNodes )
				delete[] mLeafNodes;
			if ( mBuffer )
				delete[] mBuffer;
		}
	};
}

#endif // __DD_RAY_CACHE_ENTRY_CELL__
